"""Models related to the news section in Tibia.com."""
import datetime
import re
import urllib.parse
from typing import List, Optional

from tibiapy import abc
from tibiapy.enums import NewsCategory, NewsType
from tibiapy.errors import InvalidContent
from tibiapy.utils import get_tibia_url, parse_form_data, parse_tibia_date, parse_tibiacom_content, \
    parse_tibiacom_tables, try_enum

__all__ = (
    "News",
    "NewsArchive",
    "NewsEntry",
)

ICON_PATTERN = re.compile(r"newsicon_([^_]+)_(?:small|big)")


class NewsArchive(abc.Serializable):
    """Represents the news archive.

    .. versionadded:: 5.0.0

    Attributes
    ----------
    start_date: :class:`datetime.date`
        The start date to show news for.
    end_date: :class:`datetime.date`
        The end date to show news for.
    types: :class:`list` of :class:`NewsType`
        The type of news to show.
    categories: :class:`list` of :class:`NewsCategory`.
        The categories to show.
    entries: :class:`list` of :class:`NewsEntry`
        The news matching the provided parameters.
    """

    __slots__ = (
        "start_date",
        "end_date",
        "types",
        "categories",
        "entries",
    )

    def __init__(self, **kwargs):
        self.start_date: datetime.date = kwargs.get("start_date")
        self.end_date: datetime.date = kwargs.get("end_date")
        self.types: List[NewsType] = kwargs.get("types", [])
        self.categories: List[NewsCategory] = kwargs.get("categories", [])
        self.entries: List[NewsEntry] = kwargs.get("entries", [])

    def __repr__(self):
        return f"<{self.__class__.__name__} start_date={self.start_date!r} end_date={self.end_date!r}>"

    @property
    def url(self):
        return self.get_url()

    @property
    def form_data(self):
        return self.get_form_data(self.start_date, self.end_date, self.categories, self.types)

    @classmethod
    def get_form_data(cls, start_date, end_date, categories=None, types=None):
        """Get the form data attributes to search news with specific parameters.

        Parameters
        ----------
        start_date: :class:`datetime.date`
            The beginning date to search dates in.
        end_date: :class:`datetime.date`
            The end date to search dates in.
        categories: `list` of :class:`NewsCategory`
            The allowed categories to show. If left blank, all categories will be searched.
        types: `list` of :class:`NewsType`
            The allowed news types to show. if unused, all types will be searched.

        Returns
        -------
        :class:`dict`
            A dictionary with the required form data to search news in the archive.
        """
        if not categories:
            categories = list(NewsCategory)
        if not types:
            types = list(NewsType)
        data = {
            "filter_begin_day": start_date.day,
            "filter_begin_month": start_date.month,
            "filter_begin_year": start_date.year,
            "filter_end_day": end_date.day,
            "filter_end_month": end_date.month,
            "filter_end_year": end_date.year,
        }
        for category in categories:
            key = f"filter_{category.value}"
            data[key] = category.value
        if NewsType.FEATURED_ARTICLE in types:
            data["filter_article"] = "article"
        if NewsType.NEWS in types:
            data["filter_news"] = "news"
        if NewsType.NEWS_TICKER in types:
            data["filter_ticker"] = "ticker"
        return data

    @classmethod
    def get_url(cls):
        """Get the URL to Tibia.com's news archive page.

        Notes
        -----
        It is not possible to perform a search using query parameters.
        News searches can only be performed using POST requests sending the parameters as form-data.

        Returns
        -------
        :class:`str`
            The URL to the news archive page on Tibia.com.
        """
        return get_tibia_url("news", "newsarchive")

    @classmethod
    def from_content(cls, content):
        """Get a list of news from the HTML content of the news search page.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the page.

        Returns
        -------
        :class:`NewsArchive`
            The news archive with the news found.

        Raises
        ------
        InvalidContent
            If content is not the HTML of a news search's page.
        """
        try:
            parsed_content = parse_tibiacom_content(content)
            tables = parse_tibiacom_tables(parsed_content)
            if "News Archive Search" not in tables:
                raise InvalidContent("content is not from the news archive section in Tibia.com")
            form = parsed_content.find("form")
            news_archive = cls._parse_filtering(form)
            if "Search Results" in tables:
                rows = tables["Search Results"].find_all("tr", attrs={"class": ["Odd", "Even"]})
                for row in rows:
                    cols_raw = row.find_all('td')
                    if len(cols_raw) != 3:
                        continue
                    entry = cls._parse_entry(cols_raw)
                    news_archive.entries.append(entry)
            return news_archive
        except (AttributeError, IndexError, ValueError, KeyError) as e:
            raise InvalidContent("content is not from the news archive section in Tibia.com", e)

    @classmethod
    def _parse_filtering(cls, form):
        form_data = parse_form_data(form)
        start_date = datetime.date(
            int(form_data.pop("filter_begin_year")),
            int(form_data.pop("filter_begin_month")),
            int(form_data.pop("filter_begin_day")),
        )
        end_date = datetime.date(
            int(form_data.pop("filter_end_year")),
            int(form_data.pop("filter_end_month")),
            int(form_data.pop("filter_end_day")),
        )
        news_types = []
        for news_type in NewsType:
            value = form_data.pop(news_type.filter_name, None)
            if value:
                news_types.append(news_type)
        categories = []
        for category in NewsCategory:
            value = form_data.pop(category.filter_name, None)
            if value:
                categories.append(category)
        return cls(start_date=start_date, end_date=end_date, categories=categories, types=news_types)

    @classmethod
    def _parse_entry(cls, cols_raw):
        img = cols_raw[0].find('img')
        img_url = img["src"]
        category_name = ICON_PATTERN.search(img_url)
        category = try_enum(NewsCategory, category_name.group(1))
        for br in cols_raw[1].find_all("br"):
            br.replace_with("\n")
        date_str, news_type_str = cols_raw[1].text.splitlines()
        date = parse_tibia_date(date_str)
        news_type_str = news_type_str.replace('\xa0', ' ')
        news_type = try_enum(NewsType, news_type_str)
        title = cols_raw[2].text
        news_link = cols_raw[2].find('a')
        url = urllib.parse.urlparse(news_link["href"])
        query = urllib.parse.parse_qs(url.query)
        news_id = int(query["id"][0])
        return NewsEntry(news_id, title, news_type, category, date, category_icon=img_url)


class News(abc.BaseNews, abc.Serializable):
    """Represents a news entry.

    Attributes
    ----------
    id: :class:`int`
        The internal ID of the news entry.
    title: :class:`str`
        The title of the news entry.
    category: :class:`NewsCategory`
        The category this belongs to.
    category_icon: :class:`str`
        The URL of the icon corresponding to the category.
    date: :class:`datetime.date`
        The date when the news were published.
    content: :class:`str`, optional
        The raw html content of the entry.
    thread_id: :class:`int`, optional
        The thread id of the designated discussion thread for this entry.
    """

    def __init__(self, news_id, title, content, date, category, **kwargs):
        self.id: int = news_id
        self.title: str = title
        self.content: str = content
        self.date: datetime.date = date
        self.category: NewsCategory = category
        self.thread_id: Optional[int] = kwargs.get("thread_id", None)
        self.category_icon: Optional[str] = kwargs.get("category_icon")

    # id, title, category and date inherited from BaseNews.
    __slots__ = (
        "id",
        "title",
        "category",
        "category_icon",
        "date",
        "content",
        "thread_id",
    )

    @property
    def thread_url(self):
        """:class:`str`: The URL to the thread discussing this news entry, if any."""
        return abc.BaseThread.get_url(self.thread_id) if self.thread_id else None

    @classmethod
    def from_content(cls, content, news_id=0):
        """Get a news entry by its HTML content from Tibia.com.

        Notes
        -----
        Since there's no way to obtain the entry's Id from the page contents, it will always be 0.
        A news_id can be passed to set the news_id of the resulting object.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the page.
        news_id: :class:`int`, optional
            The news_id belonging to the content being parsed.

        Returns
        -------
        :class:`News`
            The news article found in the page.

        Raises
        ------
        InvalidContent
            If content is not the HTML of a news' page.
        """
        if "(no news with id " in content:
            return None
        try:
            parsed_content = parse_tibiacom_content(content)
            # Read Information from the headline
            headline = parsed_content.find("div", attrs={"class": "NewsHeadline"})
            img = headline.find('img')
            img_url = img["src"]
            category_name = ICON_PATTERN.search(img_url)
            category = try_enum(NewsCategory, category_name.group(1))
            title_div = headline.find("div", attrs={"class": "NewsHeadlineText"})
            title = title_div.text.replace('\xa0', ' ')
            date_div = headline.find("div", attrs={"class": "NewsHeadlineDate"})
            date_str = date_div.text.replace('\xa0', ' ').replace('-', '').strip()
            date = parse_tibia_date(date_str)

            # Read the page's content.
            content_table = parsed_content.find("table")
            content_row = content_table.find("td")
            content = content_row.encode_contents().decode()
            thread_id = None
            thread_div = content_table.find("div")
            if thread_div:
                news_link = thread_div.find('a')
                url = urllib.parse.urlparse(news_link["href"])
                query = urllib.parse.parse_qs(url.query)
                thread_id = int(query["threadid"][0])

            return cls(news_id, title, content, date, category, thread_id=thread_id, category_icon=img_url)
        except AttributeError:
            raise InvalidContent("content is not from the news archive section in Tibia.com")


class NewsEntry(abc.BaseNews, abc.Serializable):
    """A news entry from the news archive.

    Attributes
    ----------
    id: :class:`int`
        The internal ID of the news entry.
    title: :class:`str`
        The title of the news entry.
        News tickers have a fragment of their content as a title.
    category: :class:`NewsCategory`
        The category this belongs to.
    category_icon: :class:`str`
        The URL of the icon corresponding to the category.
    date: :class:`datetime.date`
        The date when the news were published.
    type: :class:`NewsType`
        The type of news of this list entry.
    """

    __slots__ = (
        "id",
        "title",
        "category",
        "category_icon",
        "date",
        "type",
    )

    def __init__(self, news_id, title, news_type, category, date, **kwargs):
        self.id: int = news_id
        self.title: str = title
        self.type: NewsType = news_type
        self.category: NewsCategory = category
        self.date: datetime.datetime = date
        self.category_icon: Optional[str] = kwargs.get("category_icon", None)

    def __repr__(self):
        return f"<{self.__class__.__name__} id={self.id} title={self.title!r} type={self.type!r} " \
               f"category={self.category!r} date={self.date!r}>"



import unittest

from tests.tests_tibiapy import TestCommons
from tibiapy import GuildsSection, Highscores, KillStatistics, Vocation, Category, VocationFilter, Character, \
    AccountStatus, World, \
    WorldLocation, \
    TransferType, PvpType, GuildEntry, Guild, CharacterBazaar, BazaarType, AuctionStatus, BidType, Auction


class TestNewChanges(TestCommons, unittest.TestCase):
    def test_highscore_changes(self):
        content = self.load_resource("website_changes/1_community_highscores.txt")
        highscores = Highscores.from_content(content)

        self.assertIsNone(highscores.world)
        self.assertEqual(Category.EXPERIENCE, highscores.category)
        self.assertEqual(VocationFilter.ALL, highscores.vocation)
        self.assertEqual(1000, highscores.results_count)
        self.assertEqual(1, highscores.page)
        self.assertEqual(20, highscores.total_pages)
        first_entry = highscores.entries[0]
        self.assertEqual("Goraca", first_entry.name)
        self.assertEqual(Vocation.MASTER_SORCERER, first_entry.vocation)
        self.assertEqual(1904, first_entry.level)
        self.assertEqual(114_797_203_877, first_entry.value)

    def test_character_changes(self):
        content = self.load_resource("website_changes/2_community_characters_bobeek.txt")
        character = Character.from_content(content)

        self.assertEqual("Bobeek", character.name)
        self.assertIsNotNone(character.guild_membership)
        self.assertEqual("Hill", character.guild_membership.name)
        self.assertEqual("King", character.guild_membership.rank)
        self.assertEqual("Goraca", character.married_to)
        self.assertEqual(character.guild_name, character.guild_membership.name)
        self.assertEqual(character.guild_rank, character.guild_membership.rank)
        self.assertEqual(AccountStatus.PREMIUM_ACCOUNT, character.account_status)
        self.assertEqual(1038, character.achievement_points)
        self.assertIsNone(character.deletion_date)
        self.assertIsNotNone(character.deaths)
        self.assertEqual(1, character.deaths.__len__())
        self.assertEqual(character.url, Character.get_url(character.name))
        self.assertEqual(3, len(character.other_characters))
        self.assertFalse(character.hidden)

    def test_world_changes(self):
        content = self.load_resource("website_changes/3_community_world_adra.txt")
        world = World.from_content(content)

        self.assertEqual("Adra", world.name)
        self.assertEqual("Online", world.status)
        self.assertEqual(TransferType.BLOCKED, world.transfer_type)
        self.assertEqual(PvpType.OPEN_PVP, world.pvp_type)
        self.assertEqual(WorldLocation.EUROPE, world.location)

    def test_world_guilds_changes(self):
        content = self.load_resource("website_changes/4_community_guilds_adra.txt")
        guilds_section = GuildsSection.from_content(content)

        guilds = guilds_section.entries
        self.assertEqual(12, len(guilds))
        guild = guilds[0]

        self.assertEqual("Ambassadors", guild.name)
        self.assertEqual("Adra", guild.world)
        self.assertTrue(guild.active)

    def test_guild_changes(self):
        content = self.load_resource("website_changes/5_community_guilds_world_adra_ambassadors.txt")
        guild = Guild.from_content(content)

        self.assertEqual("Ambassadors", guild.name)
        self.assertTrue(guild.open_applications)
        self.assertEqual(31, len(guild.members))
        self.assertEqual("Adra", guild.world)
        self.assertIsNone(guild.guildhall)
        leader = guild.members[0]
        self.assertEqual("Emsky Ed", leader.name)
        self.assertEqual("Leader", leader.rank)
        self.assertEqual(293, leader.level)
        self.assertEqual(Vocation.ELDER_DRUID, leader.vocation)
        self.assertIsNone(leader.title)

    def test_bazaar_current_changes(self):
        content = self.load_resource("website_changes/6_char_bazaar_current_auctions.txt")
        char_bazaar = CharacterBazaar.from_content(content)

        self.assertEqual(BazaarType.CURRENT, char_bazaar.type)
        self.assertEqual(94, char_bazaar.total_pages)
        self.assertEqual(2339, char_bazaar.results_count)
        self.assertEqual(1, char_bazaar.page)
        first_entry = char_bazaar.entries[0]
        self.assertEqual(564217, first_entry.auction_id)
        self.assertEqual("Quaak", first_entry.name)
        self.assertEqual(664, first_entry.level)
        self.assertEqual("Estela", first_entry.world)
        self.assertEqual(Vocation.ROYAL_PALADIN, first_entry.vocation)
        self.assertEqual(AuctionStatus.IN_PROGRESS, first_entry.status)
        self.assertEqual(128, first_entry.outfit.outfit_id)
        self.assertEqual(17000, first_entry.bid)
        self.assertEqual(BidType.MINIMUM, first_entry.bid_type)

    def test_bazaar_auction_changes(self):
        content = self.load_resource("website_changes/7_char_bazaar_current_auctions_auctionid_549159.txt")
        auction = Auction.from_content(content)

        self.assertEqual("Knight Sparkizyn", auction.name)
        self.assertEqual(363, auction.level)
        self.assertEqual("Quelibra", auction.world)
        self.assertEqual(Vocation.KNIGHT, auction.vocation)
        self.assertEqual(AuctionStatus.IN_PROGRESS, auction.status)
        self.assertEqual(129, auction.outfit.outfit_id)
        self.assertEqual(3801, auction.bid)
        self.assertEqual(BidType.MINIMUM, auction.bid_type)

    def test_killstatistics_changes(self):
        content = self.load_resource("website_changes/8_community_kill_statistics_world_adra.txt")
        kill_statistics = KillStatistics.from_content(content)

        self.assertIsInstance(kill_statistics, KillStatistics)
        self.assertEqual(kill_statistics.world, "Adra")
        self.assertEqual(len(kill_statistics.entries), 784)

        # players shortcurt property
        self.assertEqual(kill_statistics.players, kill_statistics.entries["players"])
        self.assertEqual(kill_statistics.players.last_day_killed, 3)
        self.assertEqual(kill_statistics.players.last_day_killed, kill_statistics.players.last_day_players_killed)
        self.assertEqual(kill_statistics.players.last_week_killed, 76)
        self.assertEqual(kill_statistics.players.last_week_killed, kill_statistics.players.last_week_players_killed)

        # demons
        demons_entry = kill_statistics.entries["demons"]
        self.assertEqual(84, demons_entry.last_day_killed)
        self.assertEqual(1, demons_entry.last_day_players_killed)
        self.assertEqual(1439, demons_entry.last_week_killed)
        self.assertEqual(1, demons_entry.last_week_players_killed)




"""Models related to the spells section in the library."""
import os
import re
import urllib.parse
from typing import Dict, List, Optional

import bs4

from tibiapy import abc, errors
from tibiapy.enums import SpellGroup, SpellSorting, SpellType, VocationSpellFilter
from tibiapy.utils import get_tibia_url, parse_form_data, parse_integer, parse_tibiacom_content, try_enum

__all__ = (
    'SpellsSection',
    'SpellEntry',
    'Spell',
    'Rune',
)

spell_name = re.compile(r"([^(]+)\(([^)]+)\)")
group_pattern = re.compile(r"(?P<group>\w+)(?: \(Secondary Group: (?P<secondary>[^)]+))?")
cooldown_pattern = re.compile(
    r"(?P<cooldown>\d+)s \(Group: (?P<group_cooldown>\d+)s(?: ,Secondary Group: (?P<secondary_group_cooldown>\d+))?")


def to_yes_no(value: Optional[bool]):
    if value is None:
        return None
    return "yes" if value else "no"


class SpellsSection(abc.Serializable):
    """The spells section in Tibia.com.

    .. versionadded:: 5.0.0

    Attributes
    ----------
    vocation: :class:`VocationSpellFilter`
        The selected vocation filter. If :obj:`None`, spells for any vocation will be shown.
    group: :class:`SpellGroup`
        The selected spell group to display. If :obj:`None`, spells for any group will be shown.
    spell_type: :class:`SpellType`
        The selected spell type to display. If :obj:`None`, spells for any type will be shown.
    premium: :class:`bool`
        The premium status to filter in. :obj:`True` to show only premium spells,
        :obj:`False` to show free account spells and :obj:`None` will show any spells.
    sort_by: :class:`SpellSorting`
        The sorting order of the displayed spells.
    entries: :class:`SpellEntry`
        The spells matching the selected filters.
    """

    __slots__ = (
        "vocation",
        "group",
        "spell_type",
        "premium",
        "sort_by",
        "entries",
    )

    def __init__(self, **kwargs):
        self.vocation: VocationSpellFilter = kwargs.get("vocation")
        self.group: SpellGroup = kwargs.get("group")
        self.spell_type: SpellType = kwargs.get("spell_type")
        self.premium: Optional[bool] = kwargs.get("premium")
        self.sort_by: SpellSorting = kwargs.get("sort_by")
        self.entries: List[SpellEntry] = kwargs.get("entries", [])

    def __repr__(self):
        return (f"<{self.__class__.__name__} vocation={self.vocation!r} group={self.group!r} "
                f"spell_type={self.spell_type!r} len(entries)={len(self.entries)}>")

    @property
    def url(self):
        return self.get_url(vocation=self.vocation, group=self.group, spell_type=self.spell_type, premium=self.premium,
                            sort=self.sort_by)

    @classmethod
    def from_content(cls, content):
        """Parse the content of the spells section.

        Parameters
        -----------
        content: :class:`str`
            The HTML content of the page.

        Returns
        ----------
        :class:`SpellsSection`
            The spells contained and the filtering information.

        Raises
        ------
        InvalidContent
            If content is not the HTML of the spells section.
        """
        try:
            parsed_content = parse_tibiacom_content(content)
            spells_table = parsed_content.find("table")
            spell_rows = spells_table.find_all("tr")
            spells_section = cls()
            for row in spell_rows[1:]:
                columns = row.find_all("td")
                spell_link = columns[0].find("a")
                url = urllib.parse.urlparse(spell_link["href"])
                query = urllib.parse.parse_qs(url.query)
                cols_text = [c.text for c in columns]
                identifier = query["spell"][0]
                match = spell_name.findall(cols_text[0])
                name, words = match[0]
                group = try_enum(SpellGroup, cols_text[1])
                spell_type = try_enum(SpellType, cols_text[2])
                level = int(cols_text[3])
                mana = parse_integer(cols_text[4], None)
                price = parse_integer(cols_text[5], 0)
                premium = "yes" in cols_text[6]
                spell = SpellEntry(name=name.strip(), words=words.strip(), spell_type=spell_type, level=level, group=group,
                                   mana=mana, premium=premium, price=price, identifier=identifier)
                spells_section.entries.append(spell)
            form = parsed_content.find("form")
            data = parse_form_data(form)
            spells_section.vocation = try_enum(VocationSpellFilter, data["vocation"])
            spells_section.group = try_enum(SpellGroup, data["group"])
            spells_section.premium = try_enum(SpellGroup, data["group"])
            spells_section.spell_type = try_enum(SpellType, data["type"])
            spells_section.sort_by = try_enum(SpellSorting, data["sort"])
            spells_section.premium = "yes" in data["premium"] if data["premium"] else None
            return spells_section
        except (AttributeError, TypeError) as e:
            raise errors.InvalidContent("content does not belong to the Spells section", e)

    @classmethod
    def get_url(cls, *, vocation=None, group=None, spell_type=None, premium=None, sort=None):
        """Get the URL to the spells section with the desired filtering parameters.

        Parameters
        ----------
        vocation: :class:`VocationSpellFilter`, optional
            The vocation to filter in spells for.
        group: :class:`SpellGroup`, optional
            The spell's primary cooldown group.
        spell_type: :class:`SpellType`, optional
            The type of spells to show.
        premium: :class:`bool`, optional
            The type of premium requirement to filter. :obj:`None` means any premium requirement.
        sort: :class:`SpellSorting`, optional
            The field to sort spells by.

        Returns
        -------
        :class:`str`
            The URL to the spells section with the provided filtering parameters.
        """
        params = {
            "vocation": vocation.value if vocation else None,
            "group": group.value if group else None,
            "type": spell_type.value if spell_type else None,
            "sort": sort.value if sort else None,
            "premium": to_yes_no(premium),
        }
        return get_tibia_url("library", "spells", **params)


class SpellEntry(abc.Serializable):
    """A spell listed on the spells section.

    .. versionadded:: 5.0.0

    Attributes
    ----------
    identifier: :class:`str`
        The internal identifier of the spell. This is used as a key for links and images.
    name: :class:`str`
        The name of the spell.
    words: :class:`str`
        The words to cast the spell.
    group: :class:`SpellGroup`
        The cooldown group of the spell.
    spell_type: :class:`SpellType`
        The type of the spell
    exp_level: :class:`int`
        The required level to cast the spell.
    mana: :class:`int`, optional.
        The mana required to use the spell. If :obj:`None`, the mana cost is variable.
    price: :class:`int`
        The price in gold coins to learn the spell.
    premium: :class:`bool`
        Whether the spell requires a premium account to learn and use it.
    """

    __slots__ = (
        "identifier",
        "name",
        "words",
        "group",
        "spell_type",
        "exp_level",
        "mana",
        "price",
        "premium",
    )

    _serializable_properties = (
        "image_url",
    )

    def __init__(self, identifier, name, words, **kwargs):
        self.identifier: str = identifier
        self.name: str = name
        self.words: str = words
        self.group: SpellGroup = kwargs.get("group")
        self.spell_type: SpellType = kwargs.get("spell_type")
        self.exp_level: int = kwargs.get("exp_level")
        self.mana: Optional[int] = kwargs.get("mana")
        self.price: int = kwargs.get("price")
        self.premium: bool = kwargs.get("premium")

    def __repr__(self):
        return f"<{self.__class__.__name__} identifier={self.identifier!r} name={self.name!r} words={self.words!r} " \
               f"group={self.group!r} spell_type={self.spell_type!r}>"

    @property
    def url(self):
        """:class:`str`: The URL to the spell."""
        return self.get_url(self.identifier)

    @property
    def image_url(self):
        """:class:`str`: The URL to this spell's image."""
        return f"https://static.tibia.com/images/library/{self.identifier}.png"

    @classmethod
    def get_url(cls, identifier):
        """Get the URL to a spell in the Tibia.com spells section.

        Parameters
        ----------
        identifier: :class:`str`
            The identifier of the spell.

        Returns
        -------
        :class:`str`
            The URL to the spell.
        """
        return get_tibia_url("library", "spells", spell=identifier.lower())


class Spell(SpellEntry):
    """A spell listed on the spells section.

    .. versionadded:: 5.0.0

    Attributes
    ----------
    name: :class:`str`
        The name of the spell.
    words: :class:`str`
        The words to cast the spell.
    group: :class:`SpellGroup`
        The cooldown group of the spell.
    spell_type: :class:`SpellType`
        The type of the spell
    exp_level: :class:`int`
        The required level to cast the spell.
    mana: :class:`int`, optional.
        The mana required to use the spell. If :obj:`None`, the mana cost is variable.
    price: :class:`int`
        The price in gold coins to learn the spell.
    premium: :class:`bool`
        Whether the spell requires a premium account to learn and use it.
    description: :class:`str`
        A description of the spells effect and history.
    vocations: :class:`list` of :class:`str`
        The vocations that can use this spell.
    cooldown: :class:`int`
        The individual cooldown of this spell in seconds.
    cooldown_group: :class:`int`, optional
        The group cooldown of this spell in seconds.
    group_secondary: :class:`str`, optional
        The secondary cooldown group of this spell, if any.
    cooldown_group_secondary: :class:`int`, optional
        The secondary cooldown of this spell in seconds.
    soul_points: :class:`int`, optional
        The number of soul points consumed by this spell. It will be :obj:`None` if not applicable.
    amount: :class:`int`, optional
        The amount of objects created by this spell. It will be :obj:`None` if not applicable.
    magic_type: :class:`str`, optional
        The type of magic of this spell. Influenced by specialized magic level attributes.
    cities: :class:`list` of :class:`str`
        The cities where this spell can be learned.
    rune: :class:`Rune`, optinal
        Information of the rune created by this spell, if applicable.
    """

    __slots__ = (
        "description",
        "vocations",
        "cooldown",
        "group_secondary",
        "cooldown_group",
        "cooldown_group_secondary",
        "soul_points",
        "amount",
        "magic_type",
        "cities",
        "rune",
    )

    def __init__(self, identifier, name, words, **kwargs):
        super().__init__(identifier, name, words, **kwargs)
        self.description: str = kwargs.get("description")
        self.vocations: List[str] = kwargs.get("vocations")
        self.cooldown: int = kwargs.get("cooldown")
        self.group_secondary: Optional[str] = kwargs.get("group_secondary")
        self.cooldown_group: int = kwargs.get("cooldown_group")
        self.cooldown_group_secondary: Optional[int] = kwargs.get("cooldown_group_secondary")
        self.soul_points: Optional[int] = kwargs.get("soul_points")
        self.amount: Optional[int] = kwargs.get("amount")
        self.magic_type: Optional[str] = kwargs.get("magic_type")
        self.cities: List[str] = kwargs.get("cities")
        self.rune: Optional[Rune] = kwargs.get("rune")

    @classmethod
    def from_content(cls, content):
        """Parse the content of a spells page.

        Parameters
        -----------
        content: :class:`str`
            The HTML content of the page.

        Returns
        ----------
        :class:`Spell`
            The spell data. If the spell doesn't exist, this will be :obj:`None`.

        Raises
        ------
        InvalidContent
            If content is not the HTML of the spells section.
        """
        parsed_content = parse_tibiacom_content(content)
        try:
            tables = parsed_content.find_all("table")
            title_table = tables[0]
            spell_table = tables[1]
            img = title_table.find("img")
            url = urllib.parse.urlparse(img["src"])
            filename = os.path.basename(url.path)
            identifier = str(filename.split(".")[0])
            next_sibling = title_table.next_sibling
            description = ""
            while next_sibling:
                if isinstance(next_sibling, bs4.Tag):
                    if next_sibling.name == "br":
                        description += "\n"
                    elif next_sibling.name == "table":
                        break
                    else:
                        description += next_sibling.text
                elif isinstance(next_sibling, bs4.NavigableString):
                    description += str(next_sibling)
                next_sibling = next_sibling.next_sibling
            spell = cls._parse_spells_table(identifier, spell_table)
            spell.description = description.strip()
            if len(tables) > 2:
                spell.rune = cls._parse_rune_table(tables[2])
            return spell
        except (TypeError, AttributeError, IndexError) as e:
            form = parsed_content.find("form")
            if form:
                data = parse_form_data(form)
                if "subtopic=spells" in data.get("__action__"):
                    return None
            raise errors.InvalidContent("content is not a spell page", e)

    @classmethod
    def _parse_rune_table(cls, table):
        """Parse the rune information table.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table containing the rune information.

        Returns
        -------
        :class:`Rune`
            The rune described in the table.
        """
        attrs = cls._parse_table_attributes(table)
        rune = Rune(name=attrs["name"], group=try_enum(SpellGroup, attrs["group"]))
        rune.vocations = [v.strip() for v in attrs["vocation"].split(",")]
        rune.magic_type = attrs.get("magic_type")
        rune.magic_level = parse_integer(attrs.get("mag_lvl"), 0)
        rune.exp_level = parse_integer(attrs.get("exp_lvl"), 0)
        rune.mana = parse_integer(attrs.get("mana"), None)
        return rune

    @classmethod
    def _parse_spells_table(cls, identifier, spell_table):
        """Parse the table containing spell information.

        Parameters
        ----------
        identifier: :class:`str`
            The identifier of the spell.
        spell_table: :class:`bs4.Tag`
            The table containing the spell information.

        Returns
        -------
        :class:`Spell`
            The spell described in the table.
        """
        attrs = cls._parse_table_attributes(spell_table)
        spell = cls(identifier, attrs["name"], attrs["formula"], premium="yes" in attrs["premium"],
                    exp_level=int(attrs["exp_lvl"]))
        spell.vocations = [s.strip() for s in attrs["vocation"].split(",")]
        spell.cities = [s.strip() for s in attrs["city"].split(",")]
        m = group_pattern.match(attrs["group"])
        groups = m.groupdict()
        spell.group = try_enum(SpellGroup, groups.get("group"))
        spell.group_secondary = groups.get("secondary")
        m = cooldown_pattern.match(attrs["cooldown"])
        cooldowns = m.groupdict()
        spell.cooldown = int(cooldowns["cooldown"])
        spell.cooldown_group = int(cooldowns["group_cooldown"])
        spell.cooldown_group_secondary = parse_integer(cooldowns.get("secondary_group_cooldown"), None)
        spell.spell_type = try_enum(SpellType, attrs["type"])
        spell.soul_points = parse_integer(attrs.get("soul_points"), None)
        spell.mana = parse_integer(attrs.get("mana"), None)
        spell.amount = parse_integer(attrs.get("amount"), None)
        spell.price = parse_integer(attrs.get("price"), 0)
        spell.magic_type = attrs.get("magic_type")
        return spell

    @classmethod
    def _parse_table_attributes(cls, table) -> Dict[str, str]:
        """Parse the attributes of a table.

        Create a dictionary where every key is the left column (cleaned up) and the value is the right column.

        Parameters
        ----------
        table: :class:`bs4.Tag`
            The table to get the attributes from.

        Returns
        -------
        :class:`dict`
            The table attributes.
        """
        spell_rows = table.find_all("tr")
        attrs = {}
        for row in spell_rows[1:]:
            cols = row.find_all("td")
            cols_text = [c.text for c in cols]
            clean_name = cols_text[0].replace(":", "").replace(" ", "_").lower().strip()
            value = cols_text[1]
            attrs[clean_name] = value
        return attrs


class Rune(abc.Serializable):
    """Information about runes created by spells.

    .. versionadded:: 5.0.0

    Attributes
    ----------
    name: :class:`str`
        The name of the rune.
    vocations: :class:`list` of :class:`str`
        The vocations that can use this rune.
    group: :class:`SpellGroup`
        The cooldown group of the rune.
    exp_level: :class:`int`
        The experience level required to use the rune.
    mana: :class:`int`, optional
        The amount of mana needed to use the rune. It will be :obj:`None` if not applicable.
    magic_level: :class:`int`
        The magic level required to use the rune.
    magic_type: :class:`str`, optional
        The type of magic of this rune. Influenced by specialized magic level attributes.
    """

    __slots__ = (
        'name',
        'vocations',
        'group',
        'exp_level',
        'mana',
        'magic_level',
        'magic_type',
    )

    def __init__(self, **kwargs):
        self.name: str = kwargs.get("name")
        self.vocations: List[str] = kwargs.get("vocations")
        self.group: SpellGroup = kwargs.get("group")
        self.exp_level: int = kwargs.get("exp_level")
        self.mana: Optional[int] = kwargs.get("mana")
        self.magic_level: int = kwargs.get("magic_level")
        self.magic_type: Optional[str] = kwargs.get("magic_type")



"""Models related to the houses section in Tibia.com."""
import datetime
import re
from typing import Dict, List, Optional

from tibiapy import abc
from tibiapy.enums import HouseOrder, HouseStatus, HouseType, Sex
from tibiapy.errors import InvalidContent
from tibiapy.utils import get_tibia_url, parse_tibia_datetime, parse_tibia_money, \
    parse_tibiacom_content, parse_tibiacom_tables, try_date, try_datetime, try_enum

__all__ = (
    "HousesSection",
    "House",
    "CharacterHouse",
    "GuildHouse",
    "HouseEntry",
)

id_regex = re.compile(r'house_(\d+)\.')
bed_regex = re.compile(r'This (?P<type>\w+) can have up to (?P<beds>[\d-]+) bed')
info_regex = \
    re.compile(r'The house has a size of (?P<size>\d+) square meter[s]?. '
               r'The monthly rent is (?P<rent>\d+k?) gold and will be debited to the bank account on (?P<world>\w+).')

rented_regex = re.compile(r'The house has been rented by (?P<owner>[^.]+)\.'
                          r' (?P<pronoun>\w+) has paid the rent until (?P<paid_until>[^.]+)\.')
transfer_regex = re.compile(r'\w+ will move out on (?P<transfer_date>[^(]+)\([^)]+\)(?: and (?P<verb>wants to|will)'
                            r' pass the house to (?P<transferee>[\w\s]+) for (?P<transfer_price>\d+) gold coin)?')
moving_regex = re.compile(r'\w+ will move out on (?P<move_date>[^(]+)')
bid_regex = \
    re.compile(r'The highest bid so far is (?P<highest_bid>\d+) gold and has been submitted by (?P<bidder>[^.]+)')
auction_regex = re.compile(r'The auction (?P<auction_state>has ended|will end) at (?P<auction_end>[^.]+).')

list_header_regex = re.compile(r'Available (?P<type>[\w\s]+) in (?P<town>[\w\s\']+) on (?P<world>\w+)')
list_auction_regex = re.compile(r'\((?P<bid>\d+) gold; (?P<time_left>\w)+ (?P<time_unit>day|hour)s? left\)')


class HousesSection(abc.Serializable):
    """Represents the house section.

    .. versionadded:: 5.0.0

    Attributes
    ----------
    world: :class:`str`
        The selected world to show houses for.
    town: :class:`str`
        The town to show houses for.
    status: :class:`HouseStatus`
        The status to show. If :obj:`None`, any status is shown.
    house_type: :class:`HouseType`
        The type of houses to show.
    order: :class:`HouseOrder`
        The ordering to use for the results.
    entries: :class:`list` of :class:`HouseEntry`
        The houses matching the filters.
    available_worlds: :class:`list` of :class:`str`
        The list of available worlds to choose from.
    available_towns: :class:`list` of :class:`str`
        The list of available towns to choose from.
    """

    __slots__ = (
        "world",
        "town",
        "status",
        "house_type",
        "order",
        "entries",
        "available_worlds",
        "available_towns",
    )

    def __init__(self, **kwargs):
        self.world: str = kwargs.get("world")
        self.town: str = kwargs.get("town")
        self.status: Optional[HouseStatus] = kwargs.get("status")
        self.house_type: HouseType = kwargs.get("house_type")
        self.order: HouseOrder = kwargs.get("order", HouseOrder.NAME)
        self.available_worlds: List[str] = kwargs.get("available_worlds", [])
        self.available_towns: List[str] = kwargs.get("available_towns", [])
        self.entries: List[HouseEntry] = kwargs.get("entries", [])

    def __repr__(self):
        return f"<{self.__class__.__name__} world={self.world!r} town={self.town!r} house_type={self.house_type!r}>"

    @property
    def url(self):
        """:class:`str`: Get the URL to the houses section with the current parameters."""
        return self.get_url(self.world, self.town, self.house_type, self.status, self.order)

    @classmethod
    def _get_query_params(cls, world, town, house_type, status=None, order=None) -> Dict[str, str]:
        """Build the query parameters for a house search.

        Parameters
        ----------
        world: :class:`str`
            The world to search for houses in.
        town: :class:`str`
            The town to search houses in.
        house_type: :class:`HouseType`
            The type of houses to filter.
        status: :class:`HouseStatus`, optional
            The status of the house to filter. If undefined, all status will be included.
        order: :class:`HouseOrder`
            The field to order houses by.

        Returns
        -------
        :class:`dict`
            The query parameters representing for a house search.
        """
        params = {
            "world": world,
            "town": town,
            "type": house_type.plural if house_type else None,
            "status": status.value if status else None,
            "order": order.value if order else None,
        }
        return {k: v for k, v in params.items() if v is not None}

    @classmethod
    def get_url(cls, world, town, house_type, status=None, order=None):
        """Get the URL to the house list on Tibia.com with the specified filters.

        Parameters
        ----------
        world: :class:`str`
            The world to search in.
        town: :class:`str`
            The town to show results for.
        house_type: :class:`HouseType`
            The type of houses to show.
        status: :class:`HouseStatus`
            The status of the houses to show.
        order: :class:`HouseOrder`
            The sorting parameter to use for the results.

        Returns
        -------
        :class:`str`
            The URL to the list matching the parameters.
        """
        query = cls._get_query_params(world, town, house_type, status, order)
        return get_tibia_url("community", "houses", **query)

    @classmethod
    def from_content(cls, content):
        """Parse the content of a house list from Tibia.com into a list of houses.

        Parameters
        ----------
        content: :class:`str`
            The raw HTML response from the house list.

        Returns
        -------
        :class:`HouseSection`
            The houses found in the page.

        Raises
        ------
        InvalidContent`
            Content is not the house list from Tibia.com
        """
        try:
            parsed_content = parse_tibiacom_content(content)
            tables = parse_tibiacom_tables(parsed_content)
            house_results = cls()
            house_results._parse_filters(tables["House Search"])
            if len(tables) < 2:
                return house_results
            houses_table = tables[list(tables.keys())[0]]
            _, *rows = houses_table.find_all("tr")
            for row in rows[1:]:
                cols = row.find_all("td")
                if len(cols) != 5:
                    continue
                name = cols[0].text.replace('\u00a0', ' ')
                house = HouseEntry(name, house_results.world, 0, town=house_results.town,
                                   type=house_results.house_type)
                size = cols[1].text.replace('sqm', '')
                house.size = int(size)
                rent = cols[2].text.replace('gold', '')
                house.rent = parse_tibia_money(rent)
                status = cols[3].text.replace('\xa0', ' ')
                house._parse_status(status)
                id_input = cols[4].find("input", {'name': 'houseid'})
                house.id = int(id_input["value"])
                house_results.entries.append(house)
            return house_results
        except (ValueError, AttributeError, KeyError) as e:
            raise InvalidContent("content does not belong to a Tibia.com house list", e)

    def _parse_filters(self, filters_table):
        world_select = filters_table.find("select", {"name": "world"})
        for world_option in world_select.find_all("option"):
            world_name = world_option.text
            if "(" in world_name:
                continue
            self.available_worlds.append(world_name)
            if world_option.attrs.get("selected"):
                self.world = world_name
        for town_option in filters_table.find_all("input", {"name": "town"}):
            town_name = town_option.attrs.get("value")
            self.available_towns.append(town_name)
            if "checked" in town_option.attrs:
                self.town = town_name
        checked_status = filters_table.find("input", {"name": "state", "checked": True})
        if checked_status.attrs.get("value"):
            self.status = try_enum(HouseStatus, checked_status.attrs.get("value"))
        checked_type = filters_table.find("input", {"name": "type", "checked": True})
        if checked_type.attrs.get("value"):
            self.house_type = try_enum(HouseType, checked_type.attrs.get("value")[:-1])
        checked_order = filters_table.find("input", {"name": "order", "checked": True})
        if checked_order and checked_order.attrs.get("value"):
            self.order = try_enum(HouseOrder, checked_order.attrs.get("value"), HouseOrder.NAME)
        return self


class House(abc.BaseHouse, abc.HouseWithId, abc.Serializable):
    """Represents a house in a specific world.

    Attributes
    ----------
    id: :class:`int`
        The internal ID of the house. This is used on the website to identify houses.
    name: :class:`str`
        The name of the house.
    world: :class:`str`
        The name of the world the house belongs to.
    status: :class:`HouseStatus`
        The current status of the house.
    type: :class:`HouseType`
        The type of the house.
    image_url: :class:`str`
        The URL to the house's minimap image.
    beds: :class:`int`
        The number of beds the house has.
    size: :class:`int`
        The number of SQM the house has.
    rent: :class:`int`
        The monthly cost paid for the house, in gold coins.
    owner: :class:`str`
        The current owner of the house, if any.
    owner_sex: :class:`Sex`
        The sex of the owner of the house, if applicable.
    paid_until: :class:`datetime.datetime`, optional
        The date the last paid rent is due.
    transfer_date: :class:`datetime.datetime`, optional
        The date when the owner will move out of the house, if applicable.
    transferee: :class:`str`, optional
        The character who will receive the house when the owner moves, if applicable.
    transfer_price: :class:`int`
        The price that will be paid from the transferee to the owner for the house transfer.
    transfer_accepted: :class:`bool`
        Whether the house transfer has already been accepted or not.
    highest_bid: :class:`int`
        The currently highest bid on the house if it is being auctioned.
    highest_bidder: :class:`str`, optional
        The character that holds the highest bid.
    auction_end: :class:`datetime.datetime`, optional
        The date when the auction will end.
    """

    __slots__ = (
        "id",
        "name",
        "world",
        "status",
        "type",
        "image_url",
        "beds",
        "type",
        "size",
        "rent",
        "owner",
        "owner_sex",
        "paid_until",
        "transfer_date",
        "transferee",
        "transfer_price",
        "transfer_accepted",
        "highest_bid",
        "highest_bidder",
        "auction_end",
    )

    def __init__(self, name, world=None, **kwargs):
        self.id: int = kwargs.get("id", 0)
        self.name: str = name
        self.world: str = world
        self.image_url: str = kwargs.get("image_url")
        self.beds: int = kwargs.get("beds", 0)
        self.type: HouseType = try_enum(HouseType, kwargs.get("type"), HouseType.HOUSE)
        self.size: int = kwargs.get("size", 0)
        self.rent: int = kwargs.get("rent", 0)
        self.status: HouseStatus = try_enum(HouseStatus, kwargs.get("status"), None)
        self.owner: Optional[str] = kwargs.get("owner")
        self.owner_sex: Optional[str] = try_enum(Sex, kwargs.get("owner_sex"))
        self.paid_until: datetime.datetime = try_datetime(kwargs.get("paid_until"))
        self.transfer_date: datetime.datetime = try_datetime(kwargs.get("transfer_date"))
        self.transferee: Optional[str] = kwargs.get("transferee")
        self.transfer_price: int = kwargs.get("transfer_price", 0)
        self.transfer_accepted: bool = kwargs.get("transfer_accepted", False)
        self.highest_bid: int = kwargs.get("highest_bid", 0)
        self.highest_bidder: Optional[str] = kwargs.get("highest_bidder")
        self.auction_end: datetime.datetime = try_datetime(kwargs.get("auction_end"))

    # region Properties
    @property
    def owner_url(self):
        """:class:`str`: The URL to the Tibia.com page of the house's owner, if applicable."""
        return abc.BaseCharacter.get_url(self.owner) if self.owner is not None else None

    @property
    def transferee_url(self):
        """:class:`str`: The URL to the Tibia.com page of the character receiving the house, if applicable."""
        return abc.BaseCharacter.get_url(self.transferee) if self.transferee is not None else None

    @property
    def highest_bidder_url(self):
        """:class:`str`: The URL to the Tibia.com page of the character with the highest bid, if applicable."""
        return abc.BaseCharacter.get_url(self.highest_bidder) if self.highest_bidder is not None else None
    # endregion

    # region Public methods
    @classmethod
    def from_content(cls, content):
        """Parse a Tibia.com response into a House object.

        Parameters
        ----------
        content: :class:`str`
            HTML content of the page.

        Returns
        -------
        :class:`House`
            The house contained in the page, or None if the house doesn't exist.

        Raises
        ------
        InvalidContent
            If the content is not the house section on Tibia.com
        """
        parsed_content = parse_tibiacom_content(content)
        image_column, desc_column, *_ = parsed_content.find_all('td')
        if "Error" in image_column.text:
            return None
        image = image_column.find('img')
        for br in desc_column.find_all("br"):
            br.replace_with("\n")
        description = desc_column.text.replace("\u00a0", " ").replace("\n\n", "\n")
        lines = description.splitlines()
        try:
            name, beds, info, state, *_ = lines
        except ValueError:
            raise InvalidContent("content does is not from the house section of Tibia.com")

        house = cls(name.strip())
        house.image_url = image["src"]
        house.id = int(id_regex.search(house.image_url).group(1))
        m = bed_regex.search(beds)
        if m:
            if m.group("type").lower() in ["guildhall", "clanhall"]:
                house.type = HouseType.GUILDHALL
            else:
                house.type = HouseType.HOUSE
            house.beds = int(m.group("beds"))

        m = info_regex.search(info)
        if m:
            house.world = m.group("world")
            house.rent = parse_tibia_money(m.group("rent"))
            house.size = int(m.group("size"))

        house._parse_status(state)
        return house
    # endregion

    def _parse_status(self, status):
        """Parse the house's state description and applies the corresponding values.

        Parameters
        ----------
        status: :class:`str`
            Plain text string containing the current renting state of the house.
        """
        m = rented_regex.search(status)
        if m:
            self.status = HouseStatus.RENTED
            self.owner = m.group("owner")
            self.owner_sex = Sex.MALE if m.group("pronoun") == "He" else Sex.FEMALE
            self.paid_until = parse_tibia_datetime(m.group("paid_until"))
        else:
            self.status = HouseStatus.AUCTIONED

        m = transfer_regex.search(status)
        if m:
            self.transfer_date = parse_tibia_datetime(m.group("transfer_date"))
            self.transfer_accepted = m.group("verb") == "will"
            self.transferee = m.group("transferee")
            price = m.group("transfer_price")
            self.transfer_price = int(price) if price is not None else 0

        m = auction_regex.search(status)
        if m:
            self.auction_end = parse_tibia_datetime(m.group("auction_end"))

        m = bid_regex.search(status)
        if m:
            self.highest_bid = int(m.group("highest_bid"))
            self.highest_bidder = m.group("bidder")


class CharacterHouse(abc.BaseHouse, abc.HouseWithId, abc.Serializable):
    """A house owned by a character.

    Attributes
    ----------
    id: :class:`int`
        The internal ID of the house. This is used on the website to identify houses.
    name: :class:`str`
        The name of the house.
    world: :class:`str`
        The name of the world the house belongs to.
    status: :class:`HouseStatus`
        The current status of the house.
    type: :class:`HouseType`
        The type of the house.
    town: :class:`str`
        The town where the city is located in.
    owner: :class:`str`
        The owner of the house.
    paid_until_date: :class:`datetime.date`
        The date the last paid rent is due.
    """

    __slots__ = (
        "id",
        "name",
        "world",
        "status",
        "type",
        "town",
        "owner",
        "paid_until_date",
    )

    def __init__(self, _id, name, world=None, town=None, owner=None, paid_until_date=None):
        self.id = int(_id)
        self.name: str = name
        self.town: str = town
        self.world: str = world
        self.owner: str = owner
        self.paid_until_date = try_date(paid_until_date)
        self.status = HouseStatus.RENTED
        self.type = HouseType.HOUSE


class GuildHouse(abc.BaseHouse, abc.Serializable):
    """A guildhall owned by a guild.

    By limitation of Tibia.com, the ID of the guildhall is not available.

    Attributes
    ----------
    name: :class:`str`
        The name of the house.
    world: :class:`str`
        The name of the world the house belongs to.
    status: :class:`HouseStatus`
        The current status of the house.

        This is kept for compatibility with house objects and will always be :obj:`HouseStatus.RENTED`.
    type: :class:`HouseType`
        The type of the house.

        This is kept for compatibility with house objects and will always be :obj:`HouseType.GUILDHALL`.
    owner: :class:`str`
        The owner of the guildhall.
    paid_until_date: :class:`datetime.date`
        The date the last paid rent is due.
    """

    __slots__ = (
        "name",
        "world",
        "status",
        "owner",
        "paid_until_date",
    )

    def __init__(self, name, world=None, owner=None, paid_until_date=None):
        self.name: str = name
        self.world: str = world
        self.owner: str = owner
        self.paid_until_date: datetime.date = try_date(paid_until_date)
        self.status: HouseStatus = HouseStatus.RENTED
        self.type: HouseType = HouseType.GUILDHALL

    def __repr__(self):
        return "<%s name=%r>" % (self.__class__.__name__, self.name)


class HouseEntry(abc.BaseHouse, abc.HouseWithId, abc.Serializable):
    """Represents a house from the house list in Tibia.com.

    Attributes
    ----------
    id: :class:`int`
        The internal ID of the house. This is used on the website to identify houses.
    name: :class:`str`
        The name of the house.
    world: :class:`str`
        The name of the world the house belongs to.
    status: :class:`HouseStatus`
        The current status of the house.
    type: :class:`HouseType`
        The type of house.
    town: :class:`str`
        The town where the house is located.
    size: :class:`int`
        The size of the house in SQM.
    rent: :class:`int`
        The monthly cost of the house, in gold coins.
    time_left: :class:`datetime.timedelta`, optional
        The number of days or hours left until the bid ends, if it has started.
        This is not an exact measure, it is rounded to hours or days.
    highest_bid: :class:`int`, optional.
        The highest bid so far, if the auction has started.
    """

    __slots__ = (
        "id",
        "name",
        "world",
        "status",
        "type",
        "town",
        "size",
        "rent",
        "time_left",
        "highest_bid",
    )

    def __init__(self, name, world, houseid, **kwargs):
        self.name: str = name
        self.id = int(houseid)
        self.world: str = world
        self.status = try_enum(HouseStatus, kwargs.get("status"))
        self.type = try_enum(HouseType, kwargs.get("type"))
        self.town: str = kwargs.get("town")
        self.size: int = kwargs.get("size", 0)
        self.rent: int = kwargs.get("rent", 0)
        self.time_left: Optional[datetime.timedelta] = kwargs.get("time_left")
        self.highest_bid: Optional[int] = kwargs.get("highest_bid", None)

    # region Private methods
    def _parse_status(self, status):
        """
        Parse the status string found in the table and applies the corresponding values.

        Parameters
        ----------
        status: :class:`str`
            The string containing the status.
        """
        if "rented" in status:
            self.status = HouseStatus.RENTED
        else:
            m = list_auction_regex.search(status)
            if m:
                self.highest_bid = int(m.group('bid'))
                if m.group("time_unit") == "day":
                    self.time_left = datetime.timedelta(days=int(m.group("time_left")))
                else:
                    self.time_left = datetime.timedelta(hours=int(m.group("time_left")))
            self.status = HouseStatus.AUCTIONED
    # endregion



"""Models related to the leaderboard section in Tibia.com."""
import datetime
import re
from typing import List, Optional

from tibiapy import abc, errors
from tibiapy.utils import get_tibia_url, parse_form_data, parse_pagination, parse_tibia_datetime, parse_tibiacom_content

__all__ = (
    'Leaderboard',
    'LeaderboardEntry',
    'LeaderboardRotation',
)

rotation_end_pattern = re.compile(r"ends on ([^)]+)")


class Leaderboard(abc.Serializable):
    """Represents the Tibiadrome leaderboards.

    .. versionadded:: 5.0.0

    Attributes
    ----------
    world: :class:`str`
        The world this leaderboards are for.
    available_worlds: :class:`list` of :class:`str`
        The worlds available for selection.
    rotation: :class:`LeaderboardRotation`
        The rotation this leaderboards' entries are for.
    available_rotations: :class:`list` of :class:`LeaderboardRotation`
        The available rotations for selection.
    entries: list of :class:`LeaderboardEntry`
        The list of entries in this leaderboard.
    last_update: :class:`datetime.timedelta`
        How long ago was the currently displayed data updated. Only available for the current rotation.
    page: :class:`int`
        The page number being displayed.
    total_pages: :class:`int`
        The total number of pages.
    results_count: :class:`int`
        The total amount of entries in this rotation. These may be shown in another page.
    """

    __slots__ = (
        "world",
        "rotation",
        "last_update",
        "page",
        "total_pages",
        "results_count",
        "entries",
        "available_worlds",
        "available_rotations",
    )

    def __init__(self, world, rotation, **kwargs):
        self.world: str = world
        self.rotation: LeaderboardRotation = rotation
        self.available_worlds: List[str] = kwargs.get("available_worlds", [])
        self.available_rotations: List[LeaderboardRotation] = kwargs.get("available_rotations", [])
        self.entries: List[LeaderboardEntry] = kwargs.get("entries", [])
        self.last_update: Optional[datetime.timedelta] = kwargs.get("last_update")
        self.page: int = kwargs.get("page", 1)
        self.total_pages: int = kwargs.get("total_pages", 1)
        self.results_count: int = kwargs.get("results_count", 0)

    def __repr__(self):
        return f"<{self.__class__.__name__} world={self.world!r} rotation={self.rotation!r} " \
               f"page={self.page!r} total_pages={self.total_pages!r} results_count={self.results_count!r}>"

    @property
    def url(self):
        """:class:`str`: The URL to the current leaderboard."""
        return self.get_url(self.world, self.rotation.rotation_id, self.page)

    @property
    def previous_page_url(self):
        """:class:`str`: The URL to the previous page of the current leaderboard results, if there's any."""
        return self.get_page_url(self.page - 1) if self.page > 1 else None

    @property
    def next_page_url(self):
        """:class:`str`: The URL to the next page of the current leaderboard results, if there's any."""
        return self.get_page_url(self.page + 1) if self.page < self.total_pages else None

    def get_page_url(self, page):
        """Get the URL of the leaderboard at a specific page, with the current date parameters.

        Parameters
        ----------
        page: :class:`int`
            The desired page.

        Returns
        -------
        :class:`str`
            The URL to the desired page.

        Raises
        ------
        ValueError
            If the specified page is zer or less.
        """
        return self.get_url(self.world, self.rotation.rotation_id, page)

    @classmethod
    def get_url(cls, world, rotation_id=None, page=1):
        """Get the URL to the leaderboards of a world.

        Parameters
        ----------
        world: :class:`str`
            The desired world.
        rotation_id: :class:`int`
            The ID of the desired rotation. If undefined, the current rotation is shown.
        page: :class:`int`
            The desired page. By default, the first page is returned.

        Returns
        -------
        :class:`str`
            The URL to the leaderboard with the desired parameters.

        Raises
        ------
        ValueError
            If the specified page is zer or less.
        """
        if page <= 0:
            raise ValueError("page must be 1 or greater")
        return get_tibia_url("community", "leaderboards", world=world, rotation=rotation_id, currentpage=page)

    @classmethod
    def from_content(cls, content):
        """Parse the content of the leaderboards page.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the leaderboards page.

        Returns
        -------
        :class:`Leaderboard`
            The ledaerboard if found.
        """
        try:
            parsed_content = parse_tibiacom_content(content)
            tables = parsed_content.find_all("table", {"class": "TableContent"})
            form = parsed_content.find("form")
            data = parse_form_data(form, include_options=True)
            current_world = data["world"]
            current_rotation = None
            rotations = []
            for label, value in data["__options__"]["rotation"].items():
                current = False
                if "Current" in label:
                    label = "".join(rotation_end_pattern.findall(label))
                    current = True
                rotation_end = parse_tibia_datetime(label)
                rotation = LeaderboardRotation(int(value), rotation_end, current)
                if value == data["rotation"]:
                    current_rotation = rotation
                rotations.append(rotation)
            leaderboard = cls(current_world, current_rotation)
            leaderboard.available_worlds = [w for w in data["__options__"]["world"].values() if w]
            leaderboard.available_rotations = rotations
            if leaderboard.rotation and leaderboard.rotation.current:
                last_update_table = tables[2]
                numbers = re.findall(r'(\d+)', last_update_table.text)
                if numbers:
                    leaderboard.last_update = datetime.timedelta(minutes=int(numbers[0]))
            leaderboard._parse_entries(tables[-1])
            pagination_block = parsed_content.find("small")
            pages, total, count = parse_pagination(pagination_block) if pagination_block else (0, 0, 0)
            leaderboard.page = pages
            leaderboard.total_pages = total
            leaderboard.results_count = count
            return leaderboard
        except (AttributeError, ValueError) as e:
            raise errors.InvalidContent("content does not belong to the leaderboards", e)

    def _parse_entries(self, entries_table):
        entries_rows = entries_table.find_all("tr", {'style': True})
        for row in entries_rows:
            columns_raw = row.find_all("td")
            cols = [c.text for c in columns_raw]
            rank, name, points = cols
            entry = LeaderboardEntry(int(rank.replace(".", "")), name, int(points))
            self.entries.append(entry)


class LeaderboardEntry(abc.BaseCharacter, abc.Serializable):
    """Represents a single leadeboard entry.

    Parameters
    ----------
    rank: :class:`int`
        The rank of this entry.
    name: :class:`str`
        The name of the character.
    drome_level: :class:`int`
        The Tibia Drome level of this entry.
    """

    __slots__ = (
        "rank",
        "name",
        "drome_level",
    )

    def __init__(self, rank, name, drome_level):
        self.rank: int = rank
        self.name: str = name
        self.drome_level: int = drome_level

    def __repr__(self):
        return f"<{self.__class__.__name__} rank={self.rank} name={self.name!r} drome_level={self.drome_level}>"


class LeaderboardRotation(abc.Serializable):
    """A leaderboard rotation.

    .. versionadded:: 5.0.0

    Parameters
    ----------
    rotation_id: :class:`int`
        The internal ID of the rotation.
    current: :class:`bool`
        Whether this is the currently running rotation or not.
    end_date: :class:`datetime.datetime`
        The date and time when this rotation ends.
    """

    __slots__ = (
        "rotation_id",
        "current",
        "end_date",
    )

    def __init__(self, rotation_id, end_date, current=False):
        self.rotation_id: int = rotation_id
        self.end_date: datetime.datetime = end_date
        self.current: bool = current

    def __repr__(self):
        return f"<{self.__class__.__name__} rotation_id={self.rotation_id} end_date={self.end_date!r} " \
               f"current={self.current!r}>"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return other.rotation_id == self.rotation_id
        return False



"""Models related to the worlds section in Tibia.com."""
import re
from collections import OrderedDict
from typing import List, TYPE_CHECKING

from tibiapy import abc
from tibiapy.character import OnlineCharacter
from tibiapy.enums import BattlEyeType, PvpType, TournamentWorldType, TransferType, WorldLocation
from tibiapy.errors import InvalidContent
from tibiapy.utils import (get_tibia_url, parse_integer, parse_tibia_datetime, parse_tibia_full_date,
                           parse_tibiacom_content, try_date, try_datetime, try_enum)

if TYPE_CHECKING:
    import bs4

__all__ = (
    "WorldEntry",
    "World",
    "WorldOverview",
)

record_regexp = re.compile(r'(?P<count>[\d.,]+) players \(on (?P<date>[^)]+)\)')
battleye_regexp = re.compile(r'since ([^.]+).')


class WorldEntry(abc.BaseWorld, abc.Serializable):
    """Represents a game server listed in the World Overview section.

    Attributes
    ----------
    name: :class:`str`
        The name of the world.
    status: :class:`str`
        The current status of the world.
    online_count: :class:`int`
        The number of currently online players in the world.
    location: :class:`WorldLocation`
        The physical location of the game servers.
    pvp_type: :class:`PvpType`
        The type of PvP in the world.
    transfer_type: :class:`TransferType`
        The type of transfer restrictions this world has.
    battleye_date: :class:`datetime.date`
        The date when BattlEye was added to this world.
        If this is :obj:`None` and the world is protected, it means the world was protected from the beginning.
    battleye_type: :class:`BattlEyeType`
        The type of BattlEye protection this world has.

        .. versionadded:: 4.0.0
    experimental: :class:`bool`
        Whether the world is experimental or not.
    premium_only: :class:`bool`
        Whether only premium account players are allowed to play in this server.
    tournament_world_type: :class:`TournamentWorldType`
        The type of tournament world. :obj:`None` if this is not a tournament world.
    """

    __slots__ = (
        "name",
        "status",
        "location",
        "online_count",
        "pvp_type",
        "battleye_date",
        "battleye_type",
        "experimental",
        "premium_only",
        "tournament_world_type",
        "transfer_type",
    )

    _serializable_properties = (
        "battleye_protected",
    )

    def __init__(self, name, location=None, pvp_type=None, **kwargs):
        self.name: str = name
        self.location = try_enum(WorldLocation, location)
        self.pvp_type = try_enum(PvpType, pvp_type)
        self.status: str = kwargs.get("status")
        self.online_count: int = kwargs.get("online_count", 0)
        self.transfer_type = try_enum(TransferType, kwargs.get("transfer_type", TransferType.REGULAR))
        self.battleye_date = try_date(kwargs.get("battleye_date"))
        self.battleye_type = try_enum(BattlEyeType, kwargs.get("battleye_type"), BattlEyeType.UNPROTECTED)
        self.experimental: bool = kwargs.get("experimental", False)
        self.premium_only: bool = kwargs.get("premium_only", False)
        self.tournament_world_type = try_enum(TournamentWorldType, kwargs.get("tournament_world_type"), None)

    @property
    def battleye_protected(self):
        """:class:`bool`: Whether the server is currently protected with BattlEye or not.

        .. versionchanged:: 4.0.0
            Now a calculated property instead of a field.
        """
        return self.battleye_type and self.battleye_type != BattlEyeType.UNPROTECTED

    # region Public methods
    @classmethod
    def get_list_url(cls):
        """Get the URL to the World Overview page in Tibia.com.

        Returns
        -------
        :class:`str`
            The URL to the World Overview's page.
        """
        return WorldOverview.get_url()

    @classmethod
    def list_from_content(cls, content):
        """Parse the content of the World Overview section from Tibia.com and returns only the list of worlds.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the World Overview page in Tibia.com

        Returns
        -------
        :class:`list` of :class:`WorldEntry`
            A list of the worlds and their current information.

        Raises
        ------
        InvalidContent
            If the provided content is not the HTML content of the worlds section in Tibia.com
        """
        world_overview = WorldOverview.from_content(content)
        return world_overview.worlds

    # endregion

    # region Private methods
    def _parse_additional_info(self, additional_info, tournament=False):
        if "blocked" in additional_info:
            self.transfer_type = TransferType.BLOCKED
        elif "locked" in additional_info:
            self.transfer_type = TransferType.LOCKED
        else:
            self.transfer_type = TransferType.REGULAR
        self.experimental = "experimental" in additional_info
        self.premium_only = "premium" in additional_info
        if tournament:
            if "restricted Store products" in additional_info:
                self.tournament_world_type = TournamentWorldType.RESTRICTED
            else:
                self.tournament_world_type = TournamentWorldType.REGULAR
    # endregion


class World(abc.BaseWorld, abc.Serializable):
    """Represents a Tibia game server.

    Attributes
    ----------
    name: :class:`str`
        The name of the world.
    status: :class:`str`
        The current status of the world.
    online_count: :class:`int`
        The number of currently online players in the world.
    record_count: :class:`int`
        The server's online players record.
    record_date: :class:`datetime.datetime`
        The date when the online record was achieved.
    location: :class:`WorldLocation`
        The physical location of the game servers.
    pvp_type: :class:`PvpType`
        The type of PvP in the world.
    creation_date: :class:`str`
        The month and year the world was created. In YYYY-MM format.
    transfer_type: :class:`TransferType`
        The type of transfer restrictions this world has.
    world_quest_titles: :obj:`list` of :class:`str`
        List of world quest titles the server has achieved.
    battleye_date: :class:`datetime.date`
        The date when BattlEye was added to this world.
        If this is :obj:`None` and the world is protected, it means the world was protected from the beginning.
    battleye_type: :class:`BattlEyeType`
        The type of BattlEye protection this world has.

        .. versionadded:: 4.0.0
    experimental: :class:`bool`
        Whether the world is experimental or not.
    tournament_world_type: :class:`TournamentWorldType`
        The type of tournament world. :obj:`None` if this is not a tournament world.
    online_players: :obj:`list` of :class:`OnlineCharacter`.
        A list of characters currently online in the server.
    premium_only: :class:`bool`
        Whether only premium account players are allowed to play in this server.
    """

    __slots__ = (
        "name",
        "status",
        "location",
        "pvp_type",
        "battleye_date",
        "battleye_type",
        "experimental",
        "premium_only",
        "tournament_world_type",
        "transfer_type",
        "online_count",
        "record_count",
        "record_date",
        "creation_date",
        "world_quest_titles",
        "online_players",
    )

    _serializable_properties = (
        "battleye_protected",
    )

    def __init__(self, name, location=None, pvp_type=None, **kwargs):
        self.name: str = name
        self.location = try_enum(WorldLocation, location)
        self.pvp_type = try_enum(PvpType, pvp_type)
        self.status: bool = kwargs.get("status")
        self.online_count: int = kwargs.get("online_count", 0)
        self.record_count: int = kwargs.get("record_count", 0)
        self.record_date = try_datetime(kwargs.get("record_date"))
        self.creation_date: str = kwargs.get("creation_date")
        self.transfer_type = try_enum(TransferType, kwargs.get("transfer_type", TransferType.REGULAR))
        self.world_quest_titles: List[str] = kwargs.get("world_quest_titles", [])
        self.battleye_date = try_date(kwargs.get("battleye_date"))
        self.battleye_type = try_enum(BattlEyeType, kwargs.get("battleye_type"), BattlEyeType.UNPROTECTED)
        self.experimental: bool = kwargs.get("experimental", False)
        self.online_players: List[OnlineCharacter] = kwargs.get("online_players", [])
        self.premium_only: bool = kwargs.get("premium_only", False)
        self.tournament_world_type = try_enum(TournamentWorldType, kwargs.get("tournament_world_type"), None)

    # region Properties
    @property
    def battleye_protected(self):
        """:class:`bool`: Whether the server is currently protected with BattlEye or not.

        .. versionchanged:: 4.0.0
            Now a calculated property instead of a field.
        """
        return self.battleye_type and self.battleye_type != BattlEyeType.UNPROTECTED

    @property
    def creation_year(self):
        """:class:`int`: Returns the year when the world was created."""
        return int(self.creation_date.split("-")[0]) if self.creation_date else None

    @property
    def creation_month(self):
        """:class:`int`: Returns the month when the world was created."""
        return int(self.creation_date.split("-")[1])if self.creation_date else None
    # endregion

    # region Public methods
    @classmethod
    def from_content(cls, content):
        """Parse a Tibia.com response into a :class:`World`.

        Parameters
        ----------
        content: :class:`str`
            The raw HTML from the server's information page.

        Returns
        -------
        :class:`World`
            The World described in the page, or :obj:`None`.

        Raises
        ------
        InvalidContent
            If the provided content is not the HTML content of the world section in Tibia.com
        """
        parsed_content = parse_tibiacom_content(content)
        tables = cls._parse_tables(parsed_content)
        try:
            error = tables.get("Error")
            if error and error[0].text == "World with this name doesn't exist!":
                return None
            selected_world = parsed_content.find('option', selected=True)
            world = cls(selected_world.text)
            world._parse_world_info(tables.get("World Information", []))

            online_table = tables.get("Players Online", [])
            world.online_players = []
            for row in online_table[1:]:
                cols_raw = row.find_all('td')
                name, level, vocation = (c.text.replace('\xa0', ' ').strip() for c in cols_raw)
                world.online_players.append(OnlineCharacter(name, world.name, int(level), vocation))
        except AttributeError:
            raise InvalidContent("content is not from the world section in Tibia.com")

        return world
    # endregion

    # region Private methods
    def _parse_world_info(self, world_info_table):
        """
        Parse the World Information table from Tibia.com and adds the found values to the object.

        Parameters
        ----------
        world_info_table: :class:`list`[:class:`bs4.Tag`]
        """
        world_info = {}
        for row in world_info_table:
            cols_raw = row.find_all('td')
            cols = [ele.text.strip() for ele in cols_raw]
            field, value = cols
            field = field.replace("\xa0", "_").replace(" ", "_").replace(":", "").lower()
            value = value.replace("\xa0", " ")
            world_info[field] = value
        try:
            self.online_count = parse_integer(world_info.pop("players_online"))
        except KeyError:
            self.online_count = 0
        self.location = try_enum(WorldLocation, world_info.pop("location"))
        self.pvp_type = try_enum(PvpType, world_info.pop("pvp_type"))
        self.transfer_type = try_enum(TransferType, world_info.pop("transfer_type", None), TransferType.REGULAR)
        m = record_regexp.match(world_info.pop("online_record"))
        if m:
            self.record_count = parse_integer(m.group("count"))
            self.record_date = parse_tibia_datetime(m.group("date"))
        if "world_quest_titles" in world_info:
            self.world_quest_titles = [q.strip() for q in world_info.pop("world_quest_titles").split(",")]
        if self.world_quest_titles and "currently has no title" in self.world_quest_titles[0]:
            self.world_quest_titles = []
        self.experimental = world_info.pop("game_world_type", None) == "Experimental"
        self.tournament_world_type = try_enum(TournamentWorldType, world_info.pop("tournament_world_type", None))
        self._parse_battleye_status(world_info.pop("battleye_status"))
        self.premium_only = "premium_type" in world_info

        month, year = world_info.pop("creation_date").split("/")
        month = int(month)
        year = int(year)
        if year > 90:
            year += 1900
        else:
            year += 2000
        self.creation_date = f"{year:d}-{month:02d}"

        for k, v in world_info.items():
            try:
                setattr(self, k, v)
            except AttributeError:
                pass

    def _parse_battleye_status(self, battleye_string):
        """Parse the BattlEye string and applies the results.

        Parameters
        ----------
        battleye_string: :class:`str`
            String containing the world's Battleye Status.
        """
        m = battleye_regexp.search(battleye_string)
        if m:
            self.battleye_date = parse_tibia_full_date(m.group(1))
            self.battleye_type = BattlEyeType.PROTECTED if self.battleye_date else BattlEyeType.INITIALLY_PROTECTED
        else:
            self.battleye_date = None
            self.battleye_type = BattlEyeType.UNPROTECTED

    @classmethod
    def _parse_tables(cls, parsed_content):
        """
        Parse the information tables found in a world's information page.

        Parameters
        ----------
        parsed_content: :class:`bs4.BeautifulSoup`
            A :class:`BeautifulSoup` object containing all the content.

        Returns
        -------
        :class:`OrderedDict`[:class:`str`, :class:`list`[:class:`bs4.Tag`]]
            A dictionary containing all the table rows, with the table headers as keys.
        """
        tables = parsed_content.find_all('div', attrs={'class': 'TableContainer'})
        output = OrderedDict()
        for table in tables:
            title = table.find("div", attrs={'class': 'Text'}).text
            title = title.split("[")[0].strip()
            inner_table = table.find("div", attrs={'class': 'InnerTableContainer'})
            output[title] = inner_table.find_all("tr")
        return output
    # endregion


class WorldOverview(abc.Serializable):
    """Container class for the World Overview section.

    Attributes
    ----------
    record_count: :class:`int`
        The overall player online record.
    record_date: :class:`datetime.datetime`
        The date when the record was achieved.
    worlds: :class:`list` of :class:`WorldEntry`
        List of worlds, with limited info.
    """

    __slots__ = (
        "record_count",
        "record_date",
        "worlds",
    )

    serializable_properties = ('total_online',)

    def __init__(self, **kwargs):
        self.record_count: int = kwargs.get("record_count", 0)
        self.record_date = try_datetime(kwargs.get("record_date"))
        self.worlds: List[WorldEntry] = kwargs.get("worlds", [])

    def __repr__(self):
        return f"<{self.__class__.__name__} total_online={self.total_online:d}>"

    @property
    def total_online(self):
        """:class:`int`: Total players online across all worlds."""
        return sum(w.online_count for w in self.worlds)

    @property
    def tournament_worlds(self):
        """:class:`list` of :class:`GuildMember`: List of tournament worlds.

        Note that tournament worlds are not listed when there are no active or upcoming tournaments.
        """
        return [w for w in self.worlds if w.tournament_world_type is not None]

    @property
    def regular_worlds(self):
        """:class:`list` of :class:`WorldEntry`: List of worlds that are not tournament worlds."""
        return [w for w in self.worlds if w.tournament_world_type is None]

    @classmethod
    def get_url(cls):
        """Get the URL to the World Overview page in Tibia.com.

        Returns
        -------
        :class:`str`
            The URL to the World Overview's page.
        """
        return get_tibia_url("community", "worlds")

    @classmethod
    def from_content(cls, content):
        """Parse the content of the World Overview section from Tibia.com into an object of this class.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the World Overview page in Tibia.com

        Returns
        -------
        :class:`WorldOverview`
            An instance of this class containing all the information.

        Raises
        ------
        InvalidContent
            If the provided content is not the HTML content of the worlds section in Tibia.com
        """
        parsed_content = parse_tibiacom_content(content)
        world_overview = WorldOverview()
        try:
            record_table, *tables \
                = parsed_content.find_all("table", {"class": "TableContent"})
            m = record_regexp.search(record_table.text)
            world_overview.record_count = parse_integer(m.group("count"))
            world_overview.record_date = parse_tibia_datetime(m.group("date"))
            world_overview._parse_worlds_tables(tables)
            return world_overview
        except (AttributeError, KeyError, ValueError) as e:
            raise InvalidContent("content does not belong to the World Overview section in Tibia.com", e)

    def _parse_worlds(self, world_rows, tournament=False):
        """Parse the world columns and adds the results to :py:attr:`worlds`.

        Parameters
        ----------
        world_rows: :class:`list` of :class:`bs4.Tag`
            A list containing the rows of each world.
        tournament: :class:`bool`
            Whether these are tournament worlds or not.
        """
        for world_row in world_rows:
            cols = world_row.find_all("td")
            name = cols[0].text.strip()
            status = "Online"
            online = parse_integer(cols[1].text.strip(), None)
            if online is None:
                online = 0
                status = "Offline"
            location = cols[2].text.replace("\u00a0", " ").strip()
            pvp = cols[3].text.strip()

            world = WorldEntry(name, location, pvp, online_count=online, status=status)
            # Check Battleye icon to get information
            battleye_icon = cols[4].find("span", attrs={"class": "HelperDivIndicator"})
            if battleye_icon is not None:
                m = battleye_regexp.search(battleye_icon["onmouseover"])
                if m:
                    world.battleye_date = parse_tibia_full_date(m.group(1))
                    world.battleye_type = BattlEyeType.PROTECTED if world.battleye_date else BattlEyeType.INITIALLY_PROTECTED
            additional_info = cols[5].text.strip()
            world._parse_additional_info(additional_info, tournament)
            self.worlds.append(world)

    def _parse_worlds_tables(self, tables):
        """Parse the world columns and adds the results to :py:attr:`worlds`.

        Parameters
        ----------
        world_rows: :class:`list` of :class:`bs4.Tag`
            A list containing the rows of each world.
        tournament: :class:`bool`
            Whether these are tournament worlds or not.
        """
        for title_table, worlds_table in zip(tables, tables[1:]):
            title = title_table.text.lower()
            regular_world_rows = worlds_table.find_all("tr", attrs={"class": ["Odd", "Even"]})
            self._parse_worlds(regular_world_rows, "tournament" in title)



import datetime
import unittest

import tibiapy
from tests.tests_tibiapy import TestCommons
from tibiapy import enums, utils
from tibiapy.utils import get_tibia_url, parse_integer, parse_tibia_money

TIBIA_DATETIME_CEST = "Jul 10 2018, 07:13:32 CEST"
TIBIA_DATETIME_CET = "Jan 10 2018, 07:13:32 CET"
TIBIA_DATETIME_PST = "Aug 10 2015, 23:13:32 PST"
TIBIA_DATETIME_INVALID = "13:37:00 10 Feb 2003"

TIBIA_DATE = "Jun 20 2018"
TIBIA_FULL_DATE = "July 23, 2015"
TIBIA_DATE_INVALID = "8 Nov 2018"


class TestUtils(TestCommons, unittest.TestCase):
    def test_serializable_get_item(self):
        """Testing the muliple ways to use __get__ for Serializable"""
        # Class inherits from Serializable
        world = tibiapy.World("Calmera")

        # Serializable allows accessing attributes like a dictionary
        self.assertEqual(world.name, world["name"])
        # And setting values too
        world["location"] = tibiapy.enums.WorldLocation.NORTH_AMERICA
        self.assertEqual(world.location, tibiapy.enums.WorldLocation.NORTH_AMERICA)

        # Accessing via __get__ returns KeyError instead of AttributeError to follow dictionary behaviour
        with self.assertRaises(KeyError):
            _level = world["level"]  # NOSONAR

        # Accessing an undefined attribute that is defined in __slots__ returns `None` instead of raising an exception.
        del world.location
        self.assertIsNone(world["location"])

        # New attributes can't be created by assignation
        with self.assertRaises(KeyError):
            world["custom"] = "custom value"

    def test_parse_tibia_datetime(self):
        time = utils.parse_tibia_datetime(TIBIA_DATETIME_CEST)
        self.assertIsInstance(time, datetime.datetime)
        self.assertEqual(time.month, 7)
        self.assertEqual(time.day, 10)
        self.assertEqual(time.year, 2018)
        self.assertEqual(time.hour, 5)
        self.assertEqual(time.minute, 13)
        self.assertEqual(time.second, 32)

        time = utils.parse_tibia_datetime(TIBIA_DATETIME_CET)
        self.assertIsInstance(time, datetime.datetime)
        self.assertEqual(time.month, 1)
        self.assertEqual(time.day, 10)
        self.assertEqual(time.year, 2018)
        self.assertEqual(time.hour, 6)
        self.assertEqual(time.minute, 13)
        self.assertEqual(time.second, 32)

    def test_parse_tibia_datetime_invalid_timezone(self):
        time = utils.parse_tibia_datetime(TIBIA_DATETIME_PST)
        self.assertIsNone(time)

    def test_parse_tibia_datetime_invalid_datetime(self):
        time = utils.parse_tibia_datetime(TIBIA_DATETIME_INVALID)
        self.assertIsNone(time)

    def test_parse_tibia_date(self):
        date = utils.parse_tibia_date(TIBIA_DATE)
        self.assertIsInstance(date, datetime.date)
        self.assertEqual(date.month, 6)
        self.assertEqual(date.day, 20)
        self.assertEqual(date.year, 2018)

    def test_parse_tibia_date_full(self):
        date = utils.parse_tibia_full_date(TIBIA_FULL_DATE)
        self.assertIsInstance(date, datetime.date)
        self.assertEqual(date.month, 7)
        self.assertEqual(date.day, 23)
        self.assertEqual(date.year, 2015)

    def test_parse_tibia_date_invalid(self):
        date = utils.parse_tibia_date(TIBIA_DATETIME_INVALID)
        self.assertIsNone(date)

    def test_try_date(self):
        date = utils.try_date(datetime.datetime.now())
        self.assertIsInstance(date, datetime.date)

        date = utils.try_date(datetime.date.today())
        self.assertIsInstance(date, datetime.date)
        self.assertEqual(date, datetime.date.today())

        date = utils.try_date(TIBIA_DATE)
        self.assertIsInstance(date, datetime.date)

        date = utils.try_date(TIBIA_FULL_DATE)
        self.assertIsInstance(date, datetime.date)

    def test_try_date_time(self):
        date_time = utils.try_datetime(datetime.datetime.now())
        self.assertIsInstance(date_time, datetime.datetime)

        date_time = utils.try_datetime(TIBIA_DATETIME_CEST)
        self.assertIsInstance(date_time, datetime.datetime)

    def test_parse_number_words(self):
        self.assertEqual(utils.parse_number_words("one"), 1)
        self.assertEqual(utils.parse_number_words("no"), 0)
        self.assertEqual(utils.parse_number_words("..."), 0)
        self.assertEqual(utils.parse_number_words("twenty-one"), 21)
        self.assertEqual(utils.parse_number_words("one hundred two"), 102)
        self.assertEqual(utils.parse_number_words("two thousand forty five"), 2045)

    def test_try_enum(self):
        self.assertEqual(utils.try_enum(enums.Sex, "male"), enums.Sex.MALE)
        self.assertEqual(utils.try_enum(enums.TransferType, "", enums.TransferType.REGULAR), enums.TransferType.REGULAR)
        self.assertEqual(utils.try_enum(enums.WorldLocation, enums.WorldLocation.EUROPE), enums.WorldLocation.EUROPE)
        self.assertEqual(utils.try_enum(enums.VocationFilter, 4), enums.VocationFilter.SORCERERS)
        self.assertEqual(utils.try_enum(enums.Category, "FISHING"), enums.Category.FISHING)

    def test_enum_str(self):
        self.assertEqual(str(enums.Sex.MALE), enums.Sex.MALE.value)
        self.assertEqual(enums.VocationFilter.from_name("royal paladin"), enums.VocationFilter.PALADINS)
        self.assertEqual(enums.VocationFilter.from_name("unknown"), enums.VocationFilter.ALL)
        self.assertIsNone(enums.VocationFilter.from_name("unknown", False))

    def test_parse_tibia_money(self):
        self.assertEqual(1000, parse_tibia_money("1k"))
        self.assertEqual(5000000, parse_tibia_money("5kk"))
        self.assertEqual(2500, parse_tibia_money("2.5k"))
        self.assertEqual(50, parse_tibia_money("50"))
        with self.assertRaises(ValueError):
            parse_tibia_money("abc")

    def test_parse_integer(self):
        self.assertEqual(1450, parse_integer("1.450"))
        self.assertEqual(1110, parse_integer("1,110"))
        self.assertEqual(15, parse_integer("15"))
        self.assertEqual(0, parse_integer("abc"))
        self.assertEqual(-1, parse_integer("abc", -1))

    def test_get_tibia_url(self):
        self.assertEqual("https://www.tibia.com/community/?subtopic=character&name=Galarzaa+Fidera",
                         get_tibia_url("community", "character", name="Galarzaa Fidera"))
        self.assertEqual("https://www.tibia.com/community/?subtopic=character&name=Fn%F6",
                         get_tibia_url("community", "character", name="Fnö"))

    def test_parse_pagination_collapsed_first_page(self):
        """Parsing with current page 1 out of 915"""
        content = """<td class="PageNavigation"><small><div style="float: left;"><b>» <span class="PageLink 
        FirstOrLastElement"><span class="CurrentPageLink">First Page</span></span> <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp;currentpage=2">2</a></span> 
        <span class="PageLink "><a href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp
        ;currentpage=3">3</a></span> <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp;currentpage=4">4</a></span> 
        <span class="PageLink "><a href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp
        ;currentpage=5">5</a></span> <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp;currentpage=6">6</a></span> 
        <span class="PageLink "><a href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp
        ;currentpage=7">7</a></span> ... <span class="PageLink FirstOrLastElement"><a 
        href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp;currentpage=915">Last 
        Page</a></span></b></div><div style="float: right;"><b>» Results: 22874</b></div></small></td>"""
        parsed_content = utils.parse_tibiacom_content(content, builder="html5lib")
        page, total_pages, results_count = utils.parse_pagination(parsed_content)
        self.assertEqual(1, page)
        self.assertEqual(915, total_pages)
        self.assertEqual(22874, results_count)

    def test_parse_pagination_collapse_second_page(self):
        """Parsing with current page 2 out of 928"""
        content = """<td class="PageNavigation"><small><div style="float: left;"><b>» <span class="PageLink 
        FirstOrLastElement"><a href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp
        ;currentpage=1">First Page</a></span> <span class="PageLink "><span class="CurrentPageLink">2</span></span> 
        <span class="PageLink "><a href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp
        ;currentpage=3">3</a></span> <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp;currentpage=4">4</a></span> 
        <span class="PageLink "><a href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp
        ;currentpage=5">5</a></span> <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp;currentpage=6">6</a></span> 
        <span class="PageLink "><a href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp
        ;currentpage=7">7</a></span> <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp;currentpage=8">8</a></span> ... 
        <span class="PageLink FirstOrLastElement"><a 
        href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp;currentpage=928">Last 
        Page</a></span></b></div><div style="float: right;"><b>» Results: 23197</b></div></small></td> """
        parsed_content = utils.parse_tibiacom_content(content, builder="html5lib")
        page, total_pages, results_count = utils.parse_pagination(parsed_content)
        self.assertEqual(2, page)
        self.assertEqual(928, total_pages)
        self.assertEqual(23197, results_count)

    def test_parse_pagination_collapse_last_page(self):
        """Parsing the last page out of 928"""
        content = """<td class="PageNavigation"><small><div style="float: left;"><b>» <span class="PageLink 
        FirstOrLastElement"><a href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp
        ;currentpage=1">First Page</a></span> ... <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp;currentpage=925">925</a></span> 
        <span class="PageLink "><a href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp
        ;currentpage=926">926</a></span> <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=pastcharactertrades&amp;currentpage=927">927</a></span> 
        <span class="PageLink FirstOrLastElement"><span class="CurrentPageLink">Last Page</span></span></b></div><div 
        style="float: right;"><b>» Results: 23197</b></div></small></td> """
        parsed_content = utils.parse_tibiacom_content(content, builder="html5lib")
        page, total_pages, results_count = utils.parse_pagination(parsed_content)
        self.assertEqual(928, page)
        self.assertEqual(928, total_pages)
        self.assertEqual(23197, results_count)

    def test_parse_pagination_collapsed_middle(self):
        """Parsing page 300 out of 503"""
        content = """<td class="PageNavigation"><small><div style="float: left;"><b>» <span class="PageLink 
        FirstOrLastElement"><a href="https://www.tibia.com/charactertrade/?subtopic=currentcharactertrades&amp
        ;currentpage=1">First Page</a></span> ... <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=currentcharactertrades&currentpage=297">297</a></span> 
        <span class="PageLink "><a href="https://www.tibia.com/charactertrade/?subtopic=currentcharactertrades
        &currentpage=298">298</a></span> <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=currentcharactertrades&currentpage=299">299</a></span> 
        <span class="PageLink "><span class="CurrentPageLink">300</span></span> <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=currentcharactertrades&currentpage=301">301</a></span> 
        <span class="PageLink "><a href="https://www.tibia.com/charactertrade/?subtopic=currentcharactertrades
        &currentpage=302">302</a></span> <span class="PageLink "><a 
        href="https://www.tibia.com/charactertrade/?subtopic=currentcharactertrades&currentpage=303">303</a></span> 
        ... <span class="PageLink FirstOrLastElement"><a 
        href="https://www.tibia.com/charactertrade/?subtopic=currentcharactertrades&currentpage=503">Last 
        Page</a></span></b></div><div style="float: right;"><b>» Results: 12568</b></div></small></td> """
        parsed_content = utils.parse_tibiacom_content(content, builder="html5lib")
        page, total_pages, results_count = utils.parse_pagination(parsed_content)
        self.assertEqual(300, page)
        self.assertEqual(503, total_pages)
        self.assertEqual(12568, results_count)

    def parse_parse_pagination_not_collapsed_first_page(self):
        """Parsing first page with page numbers not collapsed"""
        content = """<small><div style="float: left;"><b>» Pages: <span 
        class="PageLink "><span class="CurrentPageLink">1</span></span> <span class="PageLink "><a 
        class="CipAjaxLink" ajaxcip="true" ajaxcip_datatype="Container" 
        href="https://www.tibia.com/charactertrade/ajax_getcharacterdata.php?auctionid=29122&amp;type=0&amp
        ;currentpage=2">2</a></span> <span class="PageLink "><a class="CipAjaxLink" ajaxcip="true" 
        ajaxcip_datatype="Container" href="https://www.tibia.com/charactertrade/ajax_getcharacterdata.php?auctionid
        =29122&amp;type=0&amp;currentpage=3">3</a></span> <span class="PageLink "><a class="CipAjaxLink" 
        ajaxcip="true" ajaxcip_datatype="Container" 
        href="https://www.tibia.com/charactertrade/ajax_getcharacterdata.php?auctionid=29122&amp;type=0&amp
        ;currentpage=4">4</a></span> <span class="PageLink "><a class="CipAjaxLink" ajaxcip="true" 
        ajaxcip_datatype="Container" href="https://www.tibia.com/charactertrade/ajax_getcharacterdata.php?auctionid
        =29122&amp;type=0&amp;currentpage=5">5</a></span> <span class="PageLink "><a class="CipAjaxLink" 
        ajaxcip="true" ajaxcip_datatype="Container" 
        href="https://www.tibia.com/charactertrade/ajax_getcharacterdata.php?auctionid=29122&amp;type=0&amp
        ;currentpage=6">6</a></span> <span class="PageLink "><a class="CipAjaxLink" ajaxcip="true" 
        ajaxcip_datatype="Container" href="https://www.tibia.com/charactertrade/ajax_getcharacterdata.php?auctionid
        =29122&amp;type=0&amp;currentpage=7">7</a></span> <span class="PageLink "><a class="CipAjaxLink" 
        ajaxcip="true" ajaxcip_datatype="Container" 
        href="https://www.tibia.com/charactertrade/ajax_getcharacterdata.php?auctionid=29122&amp;type=0&amp
        ;currentpage=8">8</a></span></b></div><div style="float: right;"><b>» Results: 567</b></div></small>"""
        parsed_content = utils.parse_tibiacom_content(content, builder="html5lib")
        page, total_pages, results_count = utils.parse_pagination(parsed_content)
        self.assertEqual(1, page)
        self.assertEqual(8, total_pages)
        self.assertEqual(567, results_count)




"""Base classes shared by various models."""
from __future__ import annotations

import abc
import datetime
import enum
import json
from collections import OrderedDict
from typing import Callable, TYPE_CHECKING

from tibiapy.utils import get_tibia_url

if TYPE_CHECKING:
    from tibiapy import PvpType, WorldLocation


class Serializable:
    """Contains methods to make a class convertible to JSON.

    Only attributes defined in ``__slots__`` will be serialized.

    .. note::
        | There's no way to convert JSON strings back to their original object.
        | Attempting to do so may result in data loss.
    """

    _serializable_properties = ()
    """:class:`tuple` of :class:`str`: Additional properties to serialize."""

    @classmethod
    def __slots_inherited__(cls):
        slots = []
        for base in cls.__bases__:
            try:
                # noinspection PyUnresolvedReferences
                slots.extend(base.__slots_inherited__())
            except AttributeError:
                continue
        slots.extend(getattr(cls, "__slots__", []))
        slots.extend(getattr(cls, "_serializable_properties", []))
        return tuple(OrderedDict.fromkeys(slots))

    def keys(self):
        return list(self.__slots_inherited__())

    def __getitem__(self, item):
        if item in self.keys():
            try:
                return getattr(self, item)
            except AttributeError:
                return None
        else:
            raise KeyError(item)

    def __setitem__(self, key, value):
        if key in self.keys():
            setattr(self, key, value)
        else:
            raise KeyError(key)

    @staticmethod
    def _try_dict(obj):
        try:
            if isinstance(obj, datetime.datetime):
                return obj.isoformat()
            if isinstance(obj, datetime.timedelta):
                return int(obj.total_seconds())
            if isinstance(obj, enum.Flag):
                return [str(i) for i in obj]
            if isinstance(obj, enum.Enum):
                return str(obj)
            return {k: v for k, v in dict(obj).items() if v is not None}
        except TypeError:
            return str(obj)

    def to_json(self, *, indent=None, sort_keys=False):
        """Get the object's JSON representation.

        Parameters
        ----------
        indent: :class:`int`, optional
            Number of spaces used as indentation, :obj:`None` will return the shortest possible string.
        sort_keys: :class:`bool`, optional
            Whether keys should be sorted alphabetically or preserve the order defined by the object.

        Returns
        -------
        :class:`str`
            JSON representation of the object.
        """
        return json.dumps({k: v for k, v in dict(self).items() if v is not None}, indent=indent, sort_keys=sort_keys,
                          default=self._try_dict)


class BaseAnnouncement(metaclass=abc.ABCMeta):
    """Base class for all announcement classes.

    Implement common properties and methods for announcements.

    The following implement this class:

    - :class:`.ForumAnnouncement`
    - :class:`.AnnouncementEntry`

    Attributes
    ----------
    announcement_id: :class:`int`
        The ID of the announcement.
    """

    announcement_id: int
    __slots__ = (
        "announcement_id",
    )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return other.announcement_id == self.announcement_id
        return False

    @property
    def url(self):
        """:class:`str` Get the URL to this announcement."""
        return self.get_url(self.announcement_id)

    @classmethod
    def get_url(cls, announcement_id):
        """Get the URL to an announcement with a given ID.

        Parameters
        ----------
        announcement_id: :class:`int`
            The ID of the announcement

        Returns
        -------
        :class:`str`
            The URL of the announcement.
        """
        return get_tibia_url("forum", None, action="announcement", announcementid=announcement_id)


class BaseBoard(metaclass=abc.ABCMeta):
    """Base class for all board classes.

    Implements common properties and methods for boards.

    The following implement this class:

    - :class:`.ForumBoard`
    - :class:`.BoardEntry`

    Attributes
    ----------
    board_id: :class:`int`
        The ID of the board.
    """

    __slots__ = (
        "board_id",
    )

    def __eq__(self, o: object) -> bool:
        """Two boards are considered equal if their ids are equal."""
        if isinstance(o, self.__class__):
            return self.board_id == o.board_id
        return False

    def __repr__(self):
        return f"<{self.__class__.__name__} board_id={self.board_id!r}>"

    @property
    def url(self):
        """:class:`str`: The URL of this board."""
        return self.get_url(self.board_id)

    @classmethod
    def get_url(cls, board_id, page=1, age=30):
        """Get the Tibia.com URL to a board with a given id.

        Parameters
        ----------
        board_id: :class:`int`
            The ID of the board.
        page: :class:`int`
            The page to go to.
        age: :class:`int`
            The age in days of the threads to display.

        Returns
        -------
        :class:`str`
            The URL to the board.
        """
        return get_tibia_url("forum", None, action="board", boardid=board_id, pagenumber=page, threadage=age)

    @classmethod
    def get_world_boards_url(cls):
        """Get the URL to the World Boards section in Tibia.com.

        Returns
        -------
        :class:`str`:
            The URL to the World Boards.
        """
        return get_tibia_url("forum", "worldboards")

    @classmethod
    def get_trade_boards_url(cls):
        """Get the URL to the Trade Boards section in Tibia.com.

        Returns
        -------
        :class:`str`:
            The URL to the Trade Boards.
        """
        return get_tibia_url("forum", "tradeboards")

    @classmethod
    def get_community_boards_url(cls):
        """Get the URL to the Community Boards section in Tibia.com.

        Returns
        -------
        :class:`str`:
            The URL to the Community Boards.
        """
        return get_tibia_url("forum", "communityboards")

    @classmethod
    def get_support_boards_url(cls):
        """Get the URL to the Support Boards section in Tibia.com.

        Returns
        -------
        :class:`str`:
            The URL to the Support Boards.
        """
        return get_tibia_url("forum", "supportboards")


class BaseCharacter(metaclass=abc.ABCMeta):
    """Base class for all character classes.

    Implements common properties methods for characters.

    The following implement this class:

    - :class:`.Character`
    - :class:`.GuildInvite`
    - :class:`.GuildMember`
    - :class:`.HighscoresEntry`
    - :class:`.TournamentLeaderboardEntry`
    - :class:`.OnlineCharacter`
    - :class:`.OtherCharacter`
    - :class:`.AuctionEntry`
    - :class:`.Auction`

    Attributes
    ----------
    name: :class:`str`
        The name of the character.
    """

    __slots__ = (
        "name",
    )

    def __eq__(self, o: object) -> bool:
        """Two characters are considered equal if their names are equal."""
        if isinstance(o, self.__class__):
            return self.name.lower() == o.name.lower()
        return False

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r}>"

    @property
    def url(self):
        """:class:`str`: The URL of the character's information page on Tibia.com."""
        return self.get_url(self.name)

    @classmethod
    def get_url(cls, name):
        """Get the Tibia.com URL for a given character name.

        Parameters
        ------------
        name: :class:`str`
            The name of the character.

        Returns
        --------
        :class:`str`
            The URL to the character's page.
        """
        return get_tibia_url("community", "characters", name=name)


class BaseGuild(metaclass=abc.ABCMeta):
    """Base class for Guild classes.

    The following implement this class:

    - :class:`.Guild`
    - :class:`.GuildMembership`
    - :class:`.GuildEntry`

    Attributes
    ----------
    name: :class:`str`
        The name of the guild.
    """

    __slots__ = (
        "name",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r}>"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name
        return False

    @property
    def url(self):
        """:class:`str`: The URL to the guild's information page on Tibia.com."""
        return self.get_url(self.name)

    @property
    def url_wars(self):
        """:class:`str` The URL to the guild's wars page on Tibia.com.

        .. versionadded:: 3.0.0
        """
        return self.get_url_wars(self.name)

    @classmethod
    def get_url(cls, name):
        """Get the Tibia.com URL for a given guild name.

        Parameters
        ------------
        name: :class:`str`
            The name of the guild.

        Returns
        --------
        :class:`str`
            The URL to the guild's page.
        """
        return get_tibia_url("community", "guilds", page="view", GuildName=name)

    @classmethod
    def get_url_wars(cls, name):
        """Get the Tibia.com URL for the guild wars of a guild with a given name.

        .. versionadded:: 3.0.0

        Parameters
        ------------
        name: :class:`str`
            The name of the guild.

        Returns
        --------
        :class:`str`
            The URL to the guild's wars page.
        """
        return get_tibia_url("community", "guilds", page="guildwars", action="view", GuildName=name)


class BaseHouse(metaclass=abc.ABCMeta):
    """Base class for all house classes.

    The following implement this class:

    - :class:`.House`
    - :class:`.GuildHouse`
    - :class:`.CharacterHouse`
    - :class:`.HouseEntry`

    Attributes
    ----------
    name: :class:`str`
        The name of the house.
    """

    __slots__ = (
        "name",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r}>"

    def __eq__(self, o: object) -> bool:
        """Two houses are considered equal if their names are equal."""
        if isinstance(o, self.__class__):
            return self.name.lower() == o.name.lower()
        return False

    @classmethod
    def get_url(cls, house_id, world):
        """Get the Tibia.com URL for a house with the given id and world.

        Parameters
        ----------
        house_id: :class:`int`
            The internal id of the house.
        world: :class:`str`
            The world of the house.

        Returns
        -------
        The URL to the house in Tibia.com
        """
        return get_tibia_url("community", "houses", page="view", houseid=house_id, world=world)


class BaseNews(metaclass=abc.ABCMeta):
    """Base class for all news classes.

    Implements the :py:attr:`id` attribute and common properties.

    The following implement this class:

    - :class:`.News`
    - :class:`.NewsEntry`

    Attributes
    ----------
    id: :class:`int`
        The internal ID of the news entry.
    """

    __slots__ = (
        "id",
    )

    def __eq__(self, o: object) -> bool:
        """Two news articles are considered equal if their names or ids are equal."""
        if isinstance(o, self.__class__):
            return self.id == o.id
        return False

    @property
    def url(self):
        """:class:`str`: The URL to the Tibia.com page of the news entry."""
        return self.get_url(self.id)

    @classmethod
    def get_url(cls, news_id):
        """Get the Tibia.com URL for a news entry by its id.

        Parameters
        ------------
        news_id: :class:`int`
            The id of the news entry.

        Returns
        --------
        :class:`str`
            The URL to the news' page
        """
        return get_tibia_url("news", "newsarchive", id=news_id)

    @classmethod
    def get_list_url(cls):
        """Get the URL to Tibia.com's news archive page.

        Notes
        -----
        It is not possible to perform a search using query parameters.
        News searches can only be performed using POST requests sending the parameters as form-data.

        Returns
        -------
        :class:`str`
            The URL to the news archive page on Tibia.com.
        """
        return get_tibia_url("news", "newsarchive")


class BasePost(metaclass=abc.ABCMeta):
    """Base classs for post classes.

    The following implement this class:

    - :class:`.CMPost`
    - :class:`.ForumPost`
    - :class:`.LastPost`

    Attributes
    ----------
    post_id: :class:`int`
        The internal ID of the post.
    """

    __slots__ = (
        "post_id",
    )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.post_id == other.post_id
        return False

    @property
    def url(self):
        """:class:`str`: Get the URL to this specific post."""
        return self.get_url(self.post_id)

    @classmethod
    def get_url(cls, post_id):
        """Get the URL to a specific post.

        Parameters
        ----------
        post_id: :class:`int`
            The ID of the desired post.

        Returns
        -------
        :class:`str`
            The URL to the post.
        """
        return get_tibia_url("forum", None, anchor=f"post{post_id}", action="thread", postid=post_id)


class BaseThread(metaclass=abc.ABCMeta):
    """Base class for thread classes.

    The following implement this class:

    - :class:`.ThreadEntry`
    - :class:`.ForumThread`

    Attributes
    ----------
    thread_id: :class:`int`
        The internal ID of the thread.
    """

    __slots__ = (
        "thread_id",
    )

    @property
    def url(self):
        """:class:`str`: The URL to the thread in Tibia.com."""
        return self.get_url(self.thread_id)

    @classmethod
    def get_url(cls, thread_id, page=1):
        """Get the URL to a thread with a given id.

        Parameters
        ----------
        thread_id: :class:`int`
            The id of the desired thread.
        page: :class:`int`
            The desired page, by default 1.

        Returns
        -------
        :class:`str`
            The URL to the thread.
        """
        return get_tibia_url("forum", None, action="thread", threadid=thread_id, pagenumber=page)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.thread_id == other.thread_id
        return False


class BaseTournament(metaclass=abc.ABCMeta):
    """Base class for tournament classes.

    The following implement this class:

    - :class:`.TournamentEntry`
    - :class:`.Tournament`

    Attributes
    ----------
    title: :class:`str`
        The tournament's title.
    cycle: :class:`int`
        The tournament's cycle.
    """

    __slots__ = (
        "title",
        "cycle",
    )

    def __eq__(self, other):
        """Two tournaments are considered the same when they have the same title or cycle."""
        if isinstance(other, self.__class__):
            return self.title.lower() == other.title.lower() or (self.cycle > 0 and self.cycle == other.cycle)
        return False

    @property
    def url(self):
        """:class:`str`: The URL to the tournament's information page."""
        return self.get_url(self.cycle)

    @classmethod
    def get_url(cls, tournament_cycle):
        """Get the URL to a tournament's information page.

        If its cycle is provided, otherwise it shows the current tournament.

        Parameters
        ----------
        tournament_cycle: :class:`int`
            The tournament's cycle.

        Returns
        -------
        :class:`str`
            The URL to the specified tournament.
        """
        params = None
        if tournament_cycle:
            params = {
                "tournamentcycle": tournament_cycle,
                "action": "archive",
            }
        return get_tibia_url("community", "tournament", **params)


class BaseWorld(metaclass=abc.ABCMeta):
    """Base class for all World classes.

    The following implement this class:

    - :class:`.WorldEntry`
    - :class:`.World`

    Attributes
    ----------
    name: :class:`str`
        The name of the world.
    """

    name: str
    location: WorldLocation
    pvp_type: PvpType

    __slots__ = (
        "name",
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} location={self.location!r} pvp_type={self.pvp_type!r}>"

    @property
    def url(self):
        """:class:`str`: URL to the world's information page on Tibia.com."""
        return self.get_url(self.name)

    @classmethod
    def get_url(cls, name):
        """Get the URL to the World's information page on Tibia.com.

        Parameters
        ----------
        name: :class:`str`
            The name of the world.

        Returns
        -------
        :class:`str`
            The URL to the world's information page.
        """
        return get_tibia_url("community", "worlds", world=name.title())


class HouseWithId:
    """Implements the :py:attr:`id` attribute and dependant functions and properties.

    Subclasses mut also implement :class:`.BaseHouse`
    """

    name: str
    id: int
    world: str
    get_url: Callable[[int, str], str]

    def __eq__(self, o: object) -> bool:
        """Two houses are considered equal if their names or ids are equal."""
        if isinstance(o, self.__class__):
            return self.name.lower() == o.name.lower() or self.id == o.id
        return False

    @property
    def url(self):
        """:class:`str`: The URL to the Tibia.com page of the house."""
        return self.get_url(self.id, self.world) if self.id and self.world else None



import datetime
import json
import unittest

from tests.tests_tibiapy import TestCommons
from tibiapy import InvalidContent, TournamentWorldType, World, WorldOverview
from tibiapy.enums import PvpType, TransferType, WorldLocation
from tibiapy.world import WorldEntry

FILE_WORLD_FULL = "world/tibiacom_online.txt"
FILE_WORLD_FULL_OFFLINE = "world/tibiacom_offline.txt"
FILE_WORLD_TOURNAMENT = "world/tibiacom_tournament.txt"
FILE_WORLD_NOT_FOUND = "world/tibiacom_not_found.txt"
FILE_WORLD_LIST = "world/tibiacom_list_online.txt"
FILE_WORLD_LIST_OFFLINE = "world/tibiacom_list_offline.txt"


class TestWorld(TestCommons, unittest.TestCase):

    # region Tibia.com Tests
    def test_world_from_content_full(self):
        """Testing parsing a world with full information"""
        content = self.load_resource(FILE_WORLD_FULL)
        world = World.from_content(content)

        self.assertIsInstance(world, World)
        self.assertEqual(world.name, "Premia")
        self.assertEqual(world.status, "Online")
        self.assertEqual(world.record_count, 531)
        self.assertIsInstance(world.record_date, datetime.datetime)
        self.assertEqual(world.creation_date, "2002-04")
        self.assertEqual(world.creation_year, 2002)
        self.assertEqual(world.creation_month, 4)
        self.assertEqual(world.location, WorldLocation.EUROPE)
        self.assertEqual(world.pvp_type, PvpType.OPEN_PVP)
        self.assertEqual(world.transfer_type, TransferType.REGULAR)
        self.assertEqual(len(world.world_quest_titles), 4)
        self.assertTrue(world.premium_only)
        self.assertTrue(world.battleye_protected)
        self.assertEqual(world.battleye_date, datetime.date(2017, 9, 5))
        self.assertFalse(world.experimental)
        self.assertEqual(len(world.online_players), world.online_count)
        self.assertEqual(World.get_url(world.name), world.url)

        world_json_raw = world.to_json()
        world_json = json.loads(world_json_raw)
        self.assertEqual(len(world.online_players), len(world_json["online_players"]))

    def test_world_from_content_offline(self):
        """Testing parsing an offline world"""
        content = self.load_resource(FILE_WORLD_FULL_OFFLINE, )
        world = World.from_content(content)

        self.assertIsInstance(world, World)
        self.assertEqual(world.name, "Antica")
        self.assertEqual(world.status, "Offline")
        self.assertEqual(world.record_count, 1052)
        self.assertIsInstance(world.record_date, datetime.datetime)
        self.assertEqual(world.creation_date, "1997-01")
        self.assertEqual(world.creation_year, 1997)
        self.assertEqual(world.creation_month, 1)
        self.assertEqual(world.location, WorldLocation.EUROPE)
        self.assertEqual(world.pvp_type, PvpType.OPEN_PVP)
        self.assertEqual(world.transfer_type, TransferType.REGULAR)
        self.assertEqual(len(world.world_quest_titles), 5)
        self.assertFalse(world.premium_only)
        self.assertTrue(world.battleye_protected)
        self.assertEqual(world.battleye_date, datetime.date(2017, 8, 29))
        self.assertFalse(world.experimental)
        self.assertEqual(len(world.online_players), world.online_count)
        self.assertEqual(World.get_url(world.name), world.url)

    def test_world_from_content_not_found(self):
        """Testing parsing a world that doesn't exist"""
        content = self.load_resource(FILE_WORLD_NOT_FOUND)
        world = World.from_content(content)

        self.assertIsNone(world)

    def test_world_from_content_unrelated_section(self):
        """Testing parsing a world using an unrelated section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            World.from_content(content)

    def test_world_from_content_tournament(self):
        """Testing parsing a tournament world"""
        content = self.load_resource(FILE_WORLD_TOURNAMENT)
        world = World.from_content(content)

        self.assertIsInstance(world, World)
        self.assertIsInstance(world.tournament_world_type, TournamentWorldType)
        self.assertEqual(world.tournament_world_type, TournamentWorldType.REGULAR)
        self.assertEqual(world.record_count, 21)
        self.assertTrue(world.premium_only)
        self.assertFalse(world.world_quest_titles)

    # endregion

    # region Tibia.com WorldOverview Tests
    def _test_world_overview_from_content(self):
        """Testing parsing world overview"""
        content = self.load_resource(FILE_WORLD_LIST)
        world_overview = WorldOverview.from_content(content)

        self.assertIsInstance(world_overview, WorldOverview)
        self.assertEqual(WorldOverview.get_url(), WorldEntry.get_list_url())
        self.assertGreater(len(world_overview.worlds), 0)
        self.assertGreater(world_overview.total_online, 0)
        self.assertIsNotNone(world_overview.record_date)
        self.assertIsNotNone(world_overview.record_count)
        self.assertEqual(82, len(world_overview.regular_worlds))
        self.assertEqual(6, len(world_overview.tournament_worlds))

        worlds = WorldEntry.list_from_content(content)
        self.assertEqual(len(world_overview.worlds), len(worlds))

    # TODO: Enable when we have a sample again
    def _test_world_overview_from_content_offline(self):
        """Testing parsing world overview with offline worlds"""
        content = self.load_resource(FILE_WORLD_LIST_OFFLINE)
        world_overview = WorldOverview.from_content(content)

        self.assertEqual(world_overview.record_count, 64028)
        self.assertIsInstance(world_overview.record_date, datetime.datetime)
        self.assertGreater(len(world_overview.worlds), 0)
        self.assertIsInstance(world_overview.worlds[0], WorldEntry)
        self.assertIsInstance(world_overview.worlds[0].pvp_type, PvpType)
        self.assertIsInstance(world_overview.worlds[0].transfer_type, TransferType)
        self.assertIsInstance(world_overview.worlds[0].location, WorldLocation)
        self.assertIsInstance(world_overview.worlds[0].online_count, int)

    def test_world_overview_from_content_unrelated(self):
        """Testing parsing an unrealted tibia section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            WorldEntry.list_from_content(content)
    # endregion



"""Models related to the Tibia.com character page."""
import datetime
import re
import urllib.parse
from collections import OrderedDict
from typing import List, Optional, TYPE_CHECKING

from tibiapy import abc
from tibiapy.enums import AccountStatus, Sex, Vocation
from tibiapy.errors import InvalidContent
from tibiapy.house import CharacterHouse
from tibiapy.utils import (parse_popup, parse_tibia_date, parse_tibia_datetime, parse_tibiacom_content, split_list,
                           try_datetime, try_enum)

if TYPE_CHECKING:
    import bs4

# Extracts the scheduled deletion date of a character."""
deleted_regexp = re.compile(r'([^,]+), will be deleted at (.*)')
# Extracts the death's level and killers.
death_regexp = re.compile(r'Level (?P<level>\d+) by (?P<killers>.*)\.</td>')
# From the killers list, filters out the assists.
death_assisted = re.compile(r'(?P<killers>.+)\.<br/>Assisted by (?P<assists>.+)')
# From a killer entry, extracts the summoned creature
death_summon = re.compile(r'(?P<summon>.+) of <a[^>]+>(?P<name>[^<]+)</a>')
link_search = re.compile(r'<a[^>]+>[^<]+</a>')
# Extracts the contents of a tag
link_content = re.compile(r'>([^<]+)<')

house_regexp = re.compile(r'paid until (.*)')

title_regexp = re.compile(r'(.*)\((\d+) titles? unlocked\)')
badge_popup_regexp = re.compile(r"\$\(this\),\s+'([^']+)',\s+'([^']+)',")

__all__ = (
    "AccountBadge",
    "AccountInformation",
    "Achievement",
    "Character",
    "Death",
    "GuildMembership",
    "Killer",
    "OtherCharacter",
    "OnlineCharacter",
)


class AccountBadge(abc.Serializable):
    """A displayed account badge in the character's information.

    Attributes
    ----------
    name: :class:`str`
        The name of the badge.
    icon_url: :class:`str`
        The URL to the badge's icon.
    description: :class:`str`
        The description of the badge.
    """

    __slots__ = (
        "name",
        "icon_url",
        "description",
    )

    def __init__(self, name, icon_url, description):
        self.name: str = name
        self.icon_url: str = icon_url
        self.description: str = description

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} description={self.description!r}>"


class AccountInformation(abc.Serializable):
    """Contains the information of a character's account.

    This is only visible if the character is not marked as hidden.

    Attributes
    ----------
    created: :class:`datetime.datetime`
        The date when the account was created.
    position: :class:`str`, optional
        The special position of this account, if any.
    loyalty_title: :class:`str`, optional
        The loyalty title of the account, if any.
    """

    __slots__ = (
        "created",
        "loyalty_title",
        "position",
    )

    def __init__(self, created, loyalty_title=None, position=None):
        self.created: datetime.datetime = try_datetime(created)
        self.loyalty_title: Optional[str] = loyalty_title
        self.position: Optional[str] = position

    def __repr__(self):
        return f"<{self.__class__.__name__} created={self.created!r} loyalty_title={self.loyalty_title!r}>"


class Achievement(abc.Serializable):
    """Represents an achievement listed on a character's page.

    Attributes
    ----------
    name: :class:`str`
        The name of the achievement.
    grade: :class:`int`
        The grade of the achievement, also known as stars.
    secret: :class:`bool`
        Whether the achievement is secret or not.
    """

    __slots__ = (
        "name",
        "grade",
        "secret",
    )

    def __init__(self, name, grade, secret=False):
        self.name: str = name
        self.grade = int(grade)
        self.secret: bool = secret

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} grade={self.grade} secret={self.secret}>"


class Character(abc.BaseCharacter, abc.Serializable):
    """A full character from Tibia.com, obtained from its character page.

    Attributes
    ----------
    name: :class:`str`
        The name of the character.
    traded: :class:`bool`
        If the character was traded in the last 6 months.
    deletion_date: :class:`datetime.datetime`, optional
        The date when the character will be deleted if it is scheduled for deletion. Will be :obj:`None` otherwise.
    former_names: :class:`list` of :class:`str`
        Previous names of the character.
    title: :class:`str`, optional
        The character's selected title, if any.
    unlocked_titles: :class:`int`
        The number of titles the character has unlocked.
    sex: :class:`Sex`
        The character's sex.
    vocation: :class:`Vocation`
        The character's vocation.
    level: :class:`int`
        The character's level.
    achievement_points: :class:`int`
        The total of achievement points the character has.
    world: :class:`str`
        The character's current world.
    former_world: :class:`str`, optional
        The previous world the character was in, in the last 6 months.
    residence: :class:`str`
        The current hometown of the character.
    married_to: :class:`str`, optional
        The name of the character's spouse. It will be :obj:`None` if not married.
    houses: :class:`list` of :class:`CharacterHouse`
        The houses currently owned by the character.
    guild_membership: :class:`GuildMembership`, optional
        The guild the character is a member of. It will be :obj:`None` if the character is not in a guild.
    last_login: :class:`datetime.datetime`, optional
        The last time the character logged in. It will be :obj:`None` if the character has never logged in.
    position: :class:`str`, optional
        The position of the character (e.g. CipSoft Member), if any.
    comment: :class:`str`, optional
        The displayed comment.
    account_status: :class:`AccountStatus`
        Whether the character's account is Premium or Free.
    account_badges: :class:`list` of :class:`AccountBadge`
        The displayed account badges.
    achievements: :class:`list` of :class:`Achievement`
        The achievements chosen to be displayed.
    deaths: :class:`list` of :class:`Death`
        The character's recent deaths.
    deaths_truncated: :class:`bool`
        Whether the character's deaths are truncated or not.

        In some cases, there are more deaths in the last 30 days than what can be displayed.
    account_information: :class:`AccountInformation`, optional
        The character's account information. If the character is hidden, this will be :obj:`None`.
    other_characters: :class:`list` of :class:`OtherCharacter`
        Other characters in the same account.

        It will be empty if the character is hidden, otherwise, it will contain at least the character itself.
    """

    __slots__ = (
        "name",
        "former_names",
        "traded",
        "sex",
        "title",
        "unlocked_titles",
        "vocation",
        "level",
        "achievement_points",
        "world",
        "former_world",
        "residence",
        "married_to",
        "houses",
        "guild_membership",
        "last_login",
        "account_status",
        "position",
        "comment",
        "account_badges",
        "achievements",
        "deaths",
        "deaths_truncated",
        "account_information",
        "other_characters",
        "deletion_date",
    )

    _serializable_properties = (
        "deleted",
        "hidden",
    )

    def __init__(self, name=None, world=None, vocation=None, level=0, sex=None, **kwargs):
        self.name: str = name
        self.traded: bool = kwargs.get("traded", False)
        self.former_names: List[str] = kwargs.get("former_names", [])
        self.title: Optional[str] = kwargs.get("title")
        self.unlocked_titles = int(kwargs.get("unlocked_titles", 0))
        self.sex = try_enum(Sex, sex)
        self.vocation = try_enum(Vocation, vocation)
        self.level = int(level)
        self.achievement_points = int(kwargs.get("achievement_points", 0))
        self.world: str = world
        self.former_world: Optional[str] = kwargs.get("former_world")
        self.residence: str = kwargs.get("residence")
        self.married_to: Optional[str] = kwargs.get("married_to")
        self.houses: List[CharacterHouse] = kwargs.get("houses", [])
        self.guild_membership: Optional[GuildMembership] = kwargs.get("guild_membership")
        self.last_login = try_datetime(kwargs.get("last_login"))
        self.account_status = try_enum(AccountStatus, kwargs.get("account_status"))
        self.position: Optional[str] = kwargs.get("position")
        self.comment: Optional[str] = kwargs.get("comment")
        self.account_badges: List[AccountBadge] = kwargs.get("account_badges", [])
        self.achievements: List[Achievement] = kwargs.get("achievements", [])
        self.deaths: List[Death] = kwargs.get("deaths", [])
        self.deaths_truncated: bool = kwargs.get("deaths", False)
        self.account_information: Optional[AccountInformation] = kwargs.get("account_information")
        self.other_characters: List[OtherCharacter] = kwargs.get("other_characters", [])
        self.deletion_date = try_datetime(kwargs.get("deletion_date"))

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} world={self.world!r} vocation={self.vocation!r} " \
               f"level={self.level} sex={self.sex!r}>"

    # region Properties
    @property
    def deleted(self) -> bool:
        """:class:`bool`: Whether the character is scheduled for deletion or not."""
        return self.deletion_date is not None

    @property
    def guild_name(self) -> Optional[str]:
        """:class:`str`, optional: The name of the guild the character belongs to, or :obj:`None`."""
        return self.guild_membership.name if self.guild_membership else None

    @property
    def guild_rank(self) -> Optional[str]:
        """:class:`str`, optional: The character's rank in the guild they belong to, or :obj:`None`."""
        return self.guild_membership.rank if self.guild_membership else None

    @property
    def guild_url(self) -> Optional[str]:
        """:class:`str`, optional: The character's rank in the guild they belong to, or :obj:`None`."""
        return abc.BaseGuild.get_url(self.guild_membership.name) if self.guild_membership else None

    @property
    def hidden(self) -> bool:
        """:class:`bool`: Whether this is a hidden character or not."""
        return len(self.other_characters) == 0

    @property
    def married_to_url(self) -> Optional[str]:
        """:class:`str`, optional: The URL to the husband/spouse information page on Tibia.com, if applicable."""
        return self.get_url(self.married_to) if self.married_to else None
    # endregion

    # region Public methods
    @classmethod
    def from_content(cls, content):
        """Create an instance of the class from the html content of the character's page.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the page.

        Returns
        -------
        :class:`Character`
            The character contained in the page, or None if the character doesn't exist

        Raises
        ------
        InvalidContent
            If content is not the HTML of a character's page.
        """
        parsed_content = parse_tibiacom_content(content)
        tables = cls._parse_tables(parsed_content)
        char = Character()
        if not tables:
            messsage_table = parsed_content.find("div", {"class": "TableContainer"})
            if messsage_table and "Could not find character" in messsage_table.text:
                return None
        if "Character Information" in tables.keys():
            char._parse_character_information(tables["Character Information"])
        else:
            raise InvalidContent("content does not contain a tibia.com character information page.")
        char._parse_achievements(tables.get("Account Achievements", []))
        if "Account Badges" in tables:
            char._parse_badges(tables["Account Badges"])
        char._parse_deaths(tables.get("Character Deaths", []))
        char._parse_account_information(tables.get("Account Information", []))
        char._parse_other_characters(tables.get("Characters", []))
        return char
    # endregion

    # region Private methods
    def _parse_account_information(self, rows):
        """Parse the character's account information.

        Parameters
        ----------
        rows: :class:`list` of :class:`bs4.Tag`, optional
            A list of all rows contained in the table.
        """
        acc_info = {}
        if not rows:
            return
        for row in rows:
            cols_raw = row.find_all('td')
            cols = [ele.text.strip() for ele in cols_raw]
            field, value = cols
            field = field.replace("\xa0", "_").replace(" ", "_").replace(":", "").lower()
            value = value.replace("\xa0", " ")
            acc_info[field] = value
        created = parse_tibia_datetime(acc_info["created"])
        loyalty_title = None if acc_info["loyalty_title"] == "(no title)" else acc_info["loyalty_title"]
        position = acc_info.get("position")
        self.account_information = AccountInformation(created, loyalty_title, position)

    def _parse_achievements(self, rows):
        """Parse the character's displayed achievements.

        Parameters
        ----------
        rows: :class:`list` of :class:`bs4.Tag`
            A list of all rows contained in the table.
        """
        for row in rows:
            cols = row.find_all('td')
            if len(cols) != 2:
                continue
            field, value = cols
            grade = str(field).count("achievement-grade-symbol")
            name = value.text.strip()
            secret_image = value.find("img")
            secret = False
            if secret_image:
                secret = True
            self.achievements.append(Achievement(name, grade, secret))

    def _parse_badges(self, rows):
        """Parse the character's displayed badges.

        Parameters
        ----------
        rows: :class:`list` of :class:`bs4.Tag`
            A list of all rows contained in the table.
        """
        row = rows[0]
        columns = row.find_all('td')
        for column in columns:
            popup_span = column.find("span", attrs={"class": "HelperDivIndicator"})
            if not popup_span:
                # Badges are visible, but none selected.
                return
            popup = parse_popup(popup_span['onmouseover'])
            name = popup[0]
            description = popup[1].text
            icon_image = column.find("img")
            icon_url = icon_image['src']
            self.account_badges.append(AccountBadge(name, icon_url, description))

    def _parse_character_information(self, rows):
        """
        Parse the character's basic information and applies the found values.

        Parameters
        ----------
        rows: :class:`list` of :class:`bs4.Tag`
            A list of all rows contained in the table.
        """
        int_rows = ["level", "achievement_points"]
        char = {}
        houses = []
        for row in rows:
            cols_raw = row.find_all('td')
            cols = [ele.text.strip() for ele in cols_raw]
            field, value = cols
            field = field.replace("\xa0", "_").replace(" ", "_").replace(":", "").lower()
            value = value.replace("\xa0", " ")
            # This is a special case cause we need to see the link
            if field == "house":
                house_text = value
                m = house_regexp.search(house_text)
                if not m:
                    continue
                paid_until = m.group(1)
                paid_until_date = parse_tibia_date(paid_until)
                house_link = cols_raw[1].find('a')
                if not house_link:
                    continue
                url = urllib.parse.urlparse(house_link["href"])
                query = urllib.parse.parse_qs(url.query)
                houses.append({"id": int(query["houseid"][0]), "name": house_link.text.strip(),
                               "town": query["town"][0], "paid_until": paid_until_date})
                continue
            if field == "guild_membership":
                guild_link = cols_raw[1].find('a')
                rank = value.split("of the")[0]
                char["guild_membership"] = GuildMembership(guild_link.text.replace("\xa0", " "), rank.strip())

                continue
            if field in int_rows:
                value = int(value)
            char[field] = value

        # If the character is deleted, the information is fouund with the name, so we must clean it
        m = deleted_regexp.match(char["name"])
        if m:
            char["name"] = m.group(1)
            char["deletion_date"] = parse_tibia_datetime(m.group(2))

        if "(traded)" in char["name"]:
            char["name"] = char["name"].replace("(traded)", "").strip()
            char["traded"] = True

        if "former_names" in char:
            former_names = [fn.strip() for fn in char["former_names"].split(",")]
            char["former_names"] = former_names

        if "never" in char["last_login"]:
            char["last_login"] = None
        else:
            char["last_login"] = parse_tibia_datetime(char["last_login"])

        m = title_regexp.match(char.get("title", ""))
        if m:
            name = m.group(1).strip()
            unlocked = int(m.group(2))
            if name == "None":
                name = None
            char["title"] = name
            char["unlocked_titles"] = unlocked

        char["vocation"] = try_enum(Vocation, char["vocation"])
        char["sex"] = try_enum(Sex, char["sex"])
        char["account_status"] = try_enum(AccountStatus, char["account_status"])

        for k, v in char.items():
            try:
                setattr(self, k, v)
            except AttributeError:
                # This means that there is a attribute in the character's information table that does not have a
                # corresponding class attribute.
                pass
        self.houses = [CharacterHouse(h["id"], h["name"], self.world, h["town"], self.name, h["paid_until"])
                       for h in houses]

    def _parse_deaths(self, rows):
        """Parse the character's recent deaths.

        Parameters
        ----------
        rows: :class:`list` of :class:`bs4.Tag`
            A list of all rows contained in the table.
        """
        for row in rows:
            cols = row.find_all('td')
            if len(cols) != 2:
                self.deaths_truncated = True
                break
            death_time_str = cols[0].text.replace("\xa0", " ").strip()
            death_time = parse_tibia_datetime(death_time_str)
            death = str(cols[1])
            death_info = death_regexp.search(death)
            if death_info:
                level = int(death_info.group("level"))
                killers_desc = death_info.group("killers")
            else:
                continue
            death = Death(self.name, level, time=death_time)
            assists_name_list = []
            # Check if the killers list contains assists
            assist_match = death_assisted.search(killers_desc)
            if assist_match:
                # Filter out assists
                killers_desc = assist_match.group("killers")
                # Split assists into a list.
                assists_desc = assist_match.group("assists")
                assists_name_list = link_search.findall(assists_desc)
            killers_name_list = split_list(killers_desc)
            for killer in killers_name_list:
                killer = killer.replace("\xa0", " ")
                killer_dict = self._parse_killer(killer)
                death.killers.append(Killer(**killer_dict))
            for assist in assists_name_list:
                # Extract names from character links in assists list.
                assist_dict = {"name": link_content.search(assist).group(1), "player": True}
                death.assists.append(Killer(**assist_dict))
            try:
                self.deaths.append(death)
            except ValueError:
                # Some pvp deaths have no level, so they are raising a ValueError, they will be ignored for now.
                continue

    @classmethod
    def _parse_killer(cls, killer):
        """Parse a killer into a dictionary.

        Parameters
        ----------
        killer: :class:`str`
            The killer's raw HTML string.

        Returns
        -------
        :class:`dict`: A dictionary containing the killer's info.
        """
        # If the killer contains a link, it is a player.
        if "href" in killer:
            m = link_content.search(killer)
            killer_dict = {"name": m.group(1), "player": True}
        else:
            killer_dict = {"name": killer, "player": False}
        # Check if it contains a summon.
        m = death_summon.search(killer)
        if m:
            killer_dict["summon"] = m.group("summon")
        return killer_dict

    def _parse_other_characters(self, rows):
        """Parse the character's other visible characters.

        Parameters
        ----------
        rows: :class:`list` of :class:`bs4.Tag`
            A list of all rows contained in the table.
        """
        for row in rows[1:]:
            cols_raw = row.find_all('td')
            cols = [ele.text.strip() for ele in cols_raw]
            if len(cols) != 4:
                continue
            name, world, status, *__ = cols
            _, *name = name.replace("\xa0", " ").split(" ")
            name = " ".join(name)
            traded = False
            if "(traded)" in name:
                name = name.replace("(traded)", "").strip()
                traded = True
            main_img = cols_raw[0].find('img')
            main = False
            if main_img and main_img['title'] == "Main Character":
                main = True
            position = None
            if "CipSoft Member" in status:
                position = "CipSoft Member"
            self.other_characters.append(OtherCharacter(name, world, "online" in status, "deleted" in status, main,
                                                        position, traded))

    @classmethod
    def _parse_tables(cls, parsed_content):
        """
        Parse the information tables contained in a character's page.

        Parameters
        ----------
        parsed_content: :class:`bs4.BeautifulSoup`
            A :class:`BeautifulSoup` object containing all the content.

        Returns
        -------
        :class:`OrderedDict`[str, :class:`list`of :class:`bs4.Tag`]
            A dictionary containing all the table rows, with the table headers as keys.
        """
        tables = parsed_content.find_all('table', attrs={"width": "100%"})
        output = OrderedDict()
        for table in tables:
            container = table.find_parent("div", {"class": "TableContainer"})
            if container:
                caption_container = container.find("div", {"class": "CaptionContainer"})
                title = caption_container.text.strip()
                offset = 0
            else:
                title = table.find("td").text.strip()
                offset = 1
            output[title] = table.find_all("tr")[offset:]
        return output
    # endregion


class Death(abc.Serializable):
    """A character's death.

    Attributes
    -----------
    name: :class:`str`
        The name of the character this death belongs to.
    level: :class:`int`
        The level at which the death occurred.
    killers: :class:`list` of :class:`Killer`
        A list of all the killers involved.
    assists: :class:`list` of :class:`Killer`
        A list of characters that were involved, without dealing damage.
    time: :class:`datetime.datetime`
        The time at which the death occurred.
    """

    __slots__ = (
        "level",
        "killers",
        "time",
        "assists",
        "name",
    )

    _serializable_properties = (
        "by_player"
    )

    def __init__(self, name=None, level=0, **kwargs):
        self.name: str = name
        self.level: int = level
        self.killers: List[Killer] = kwargs.get("killers", [])
        self.assists: List[Killer] = kwargs.get("assists", [])
        self.time: datetime.datetime = try_datetime(kwargs.get("time"))

    def __repr__(self):
        attributes = ""
        for attr in self.__slots__:
            v = getattr(self, attr)
            if isinstance(v, int) and v == 0 and not isinstance(v, bool):
                continue
            if isinstance(v, list) and len(v) == 0:
                continue
            if v is None:
                continue
            attributes += f" {attr}={v!r}"
        return f"<{self.__class__.__name__} {attributes})"

    # region Properties
    @property
    def by_player(self):
        """:class:`bool`: Whether the kill involves other characters."""
        return any([k.player and self.name != k.name for k in self.killers])

    @property
    def killer(self):
        """:class:`Killer`: The first killer in the list.

        This is usually the killer that gave the killing blow.
        """
        return self.killers[0] if self.killers else None
    # endregion


class GuildMembership(abc.BaseGuild, abc.Serializable):
    """The guild information of a character.

    Attributes
    ----------
    name: :class:`str`
        The name of the guild.
    rank: :class:`str`
        The name of the rank the member has.
    title: :class:`str`, optional
        The title of the member in the guild. This is only available for characters in the forums section.
    """

    __slots__ = (
        "name",
        "rank",
        "title",
    )

    def __init__(self, name, rank, title=None):
        self.name: str = name
        self.rank: str = rank
        self.title: Optional[str] = title

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} rank={self.rank!r}>"


class Killer(abc.Serializable):
    """Represents a killer.

    A killer can be:

    a) A creature.
    b) A character.
    c) A creature summoned by a character.

    Attributes
    -----------
    name: :class:`str`
        The name of the killer. In the case of summons, the name belongs to the owner.
    player: :class:`bool`
        Whether the killer is a player or not.
    summon: :class:`str`, optional
        The name of the summoned creature, if applicable.
    """

    __slots__ = (
        "name",
        "player",
        "summon",
    )

    def __init__(self, name, player=False, summon=None):
        self.name: str = name
        self.player: bool = player
        self.summon: Optional[str] = summon

    def __repr__(self):
        summon = f" summon={self.summon!r}" if self.summon else ""
        return f"<{self.__class__.__name__} name={self.name!r} player={self.player}{summon}>"

    @property
    def url(self):
        """:class:`str`, optional: The URL of the character’s information page on Tibia.com, if applicable."""
        return Character.get_url(self.name) if self.player else None


class OnlineCharacter(abc.BaseCharacter, abc.Serializable):
    """An online character in the world's page.

    Attributes
    ----------
    name: :class:`str`
        The name of the character.
    world: :class:`str`
        The name of the world.
    vocation: :class:`Vocation`
        The vocation of the character.
    level: :class:`int`
        The level of the character.
    """

    __slots__ = (
        "name",
        "world",
        "vocation",
        "level",
    )

    def __init__(self, name, world, level, vocation):
        self.name: str = name
        self.world: str = world
        self.level = int(level)
        self.vocation = try_enum(Vocation, vocation)

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} level={self.level} vocation={self.vocation!r}>"


class OtherCharacter(abc.BaseCharacter, abc.Serializable):
    """A character listed in the characters section of a character's page.

    These are only shown if the character is not hidden, and only characters that are not hidden are shown here.

    Attributes
    ----------
    name: :class:`str`
        The name of the character.
    world: :class:`str`
        The name of the world.
    online: :class:`bool`
        Whether the character is online or not.
    deleted: :class:`bool`
        Whether the character is scheduled for deletion or not.
    traded: :class:`bool`
        Whether the character has been traded recently or not.
    main: :class:`bool`
        Whether this is the main character or not.
    position: :class:`str`
        The character's official position, if any.
    """

    __slots__ = (
        "name",
        "world",
        "online",
        "deleted",
        "traded",
        "main",
        "position",
    )

    def __init__(self, name, world, online=False, deleted=False, main=False, position=None, traded=False):
        self.name: str = name
        self.world: str = world
        self.online: bool = online
        self.deleted: bool = deleted
        self.main: bool = main
        self.traded: bool = traded
        self.position: Optional[str] = position

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} world={self.world!r} online={self.online!r} " \
               f"main={self.main} position={self.position!r}>"



import unittest

from tests.tests_tibiapy import TestCommons
from tibiapy import KillStatistics, InvalidContent

FILE_KILL_STATISTICS_FULL = "kill_statistics/tibiacom_full.txt"
FILE_KILL_STATISTICS_EMPTY = "kill_statistics/tibiacom_empty.txt"


class TestHighscores(TestCommons, unittest.TestCase):
    # region Tibia.com Tests
    def test_kill_statistics_from_content(self):
        """Testing parsing kill statistics"""
        content = self.load_resource(FILE_KILL_STATISTICS_FULL)
        kill_statistics = KillStatistics.from_content(content)

        self.assertEqual(kill_statistics.world, "Gladera")
        self.assertEqual(len(kill_statistics.entries), 920)
        self.assertIsNotNone(kill_statistics.total)
        self.assertIsNotNone(kill_statistics.url)

        # players shortcurt property
        self.assertEqual(kill_statistics.players, kill_statistics.entries["players"])
        self.assertEqual(kill_statistics.players.last_day_killed, 2)
        self.assertEqual(kill_statistics.players.last_day_killed, kill_statistics.players.last_day_players_killed)
        self.assertEqual(kill_statistics.players.last_week_killed, 7)
        self.assertEqual(kill_statistics.players.last_week_killed, kill_statistics.players.last_week_players_killed)

        # demons
        demons_entry = kill_statistics.entries["demons"]
        self.assertEqual(2071, demons_entry.last_day_killed)
        self.assertEqual(1, demons_entry.last_day_players_killed)
        self.assertEqual(18484, demons_entry.last_week_killed)
        self.assertEqual(8, demons_entry.last_week_players_killed)

    def test_kill_statistics_from_content_empty(self):
        """Testing parsing empty kill statistics"""
        content = self.load_resource(FILE_KILL_STATISTICS_EMPTY)
        kill_statistics = KillStatistics.from_content(content)

        self.assertIsNone(kill_statistics)

    def test_kill_statistics_from_content_unrelated_section(self):
        """Testing parsing an unrelated section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            KillStatistics.from_content(content)

    # endregion
